@@ -199,8 +199,6 @@ plpgsql_check_target(PLpgSQL_checkstate *cstate, int varno, Oid *expected_typoid
 				 * Target is an element of an array
 				 */
 				int			nsubscripts;
-				Oid			arrayelemtypeid;
-				Oid			arraytypeid;
 
 				/*
 				 * To handle constructs like x[1][2] := something, we have to
@@ -226,25 +224,31 @@ plpgsql_check_target(PLpgSQL_checkstate *cstate, int varno, Oid *expected_typoid
 					target = cstate->estate->datums[arrayelem->arrayparentno];
 				} while (target->dtype == PLPGSQL_DTYPE_ARRAYELEM);
 
-				/*
-				 * If target is domain over array, reduce to base type
-				 */
+				if (expected_typoid || expected_typmod)
+				{
+					int			arraytypmod;
+					Oid			arrayelemtypeid;
+					Oid			arraytypeid;
 
-				arraytypeid = plpgsql_check__exec_get_datum_type_p(cstate->estate, target);
-				arraytypeid = getBaseType(arraytypeid);
+					plpgsql_check_target(cstate, target->dno, &arraytypeid, &arraytypmod);
 
-				arrayelemtypeid = get_element_type(arraytypeid);
+					/*
+					 * If target is domain over array, reduce to base type
+					 */
+					arraytypeid = getBaseType(arraytypeid);
+					arrayelemtypeid = get_element_type(arraytypeid);
 
-				if (!OidIsValid(arrayelemtypeid))
-					ereport(ERROR,
-							(errcode(ERRCODE_DATATYPE_MISMATCH),
-							 errmsg("subscripted object is not an array")));
+					if (!OidIsValid(arrayelemtypeid))
+						ereport(ERROR,
+								(errcode(ERRCODE_DATATYPE_MISMATCH),
+								 errmsg("subscripted object is not an array")));
 
-				if (expected_typoid)
-					*expected_typoid = arrayelemtypeid;
+					if (expected_typoid)
+						*expected_typoid = arrayelemtypeid;
 
-				if (expected_typmod)
-					*expected_typmod = ((PLpgSQL_var *) target)->datatype->atttypmod;
+					if (expected_typmod)
+						*expected_typmod = arraytypmod;
+				}
 
 				plpgsql_check_record_variable_usage(cstate, target->dno, true);
 			}
@@ -276,14 +280,22 @@ plpgsql_check_assign_to_target_type(PLpgSQL_checkstate *cstate,
 		return;
 
 	if (type_is_rowtype(value_typoid))
+	{
+		StringInfoData	str;
+
+		initStringInfo(&str);
+		appendStringInfo(&str, "cannot cast composite value of \"%s\" type to a scalar value of \"%s\" type",
+									format_type_be(value_typoid),
+									format_type_be(target_typoid));
+
 		plpgsql_check_put_error(cstate,
 					  ERRCODE_DATATYPE_MISMATCH, 0,
-					  "cannot cast composite value to a scalar type",
+					  str.data,
 					  NULL,
 					  NULL,
 					  PLPGSQL_CHECK_ERROR,
 					  0, NULL, NULL);
-
+	}
 	else if (target_typoid != value_typoid && !isnull)
 	{
 		StringInfoData	str;
@@ -313,14 +325,13 @@ plpgsql_check_assign_to_target_type(PLpgSQL_checkstate *cstate,
 		else
 		{
 			/* highly probably only performance issue */
-			if (!isnull)
-				plpgsql_check_put_error(cstate,
-							  ERRCODE_DATATYPE_MISMATCH, 0,
-							  "target type is different type than source type",
-							  str.data,
-							  "Hidden casting can be a performance issue.",
-							  PLPGSQL_CHECK_WARNING_PERFORMANCE,
-							  0, NULL, NULL);
+			plpgsql_check_put_error(cstate,
+						  ERRCODE_DATATYPE_MISMATCH, 0,
+						  "target type is different type than source type",
+						  str.data,
+						  "Hidden casting can be a performance issue.",
+						  PLPGSQL_CHECK_WARNING_PERFORMANCE,
+						  0, NULL, NULL);
 		}
 
 		pfree(str.data);
@@ -452,8 +463,7 @@ plpgsql_check_assign_tupdesc_row_or_rec(PLpgSQL_checkstate *cstate,
 		plpgsql_check_recval_release(target);
 		plpgsql_check_recval_assign_tupdesc(cstate, target, tupdesc, isnull);
 	}
-
-	else if (row != NULL && tupdesc != NULL)
+	else if (row != NULL)
 	{
 		int			td_natts = tupdesc->natts;
 		int			fnum;
@@ -608,8 +618,8 @@ plpgsql_check_recval_assign_tupdesc(PLpgSQL_checkstate *cstate, PLpgSQL_rec *rec
 
 	if (!is_null && tupdesc != NULL && !compatible_tupdescs(var_tupdesc, tupdesc))
 	{
-		int		i = 0;
-		int		j = 0;
+		int		attn1 = 0;
+		int		attn2 = 0;
 		int		target_nfields = 0;
 		int		src_nfields = 0;
 		bool	src_field_is_valid = false;
@@ -617,26 +627,26 @@ plpgsql_check_recval_assign_tupdesc(PLpgSQL_checkstate *cstate, PLpgSQL_rec *rec
 		Form_pg_attribute sattr = NULL;
 		Form_pg_attribute tattr = NULL;
 
-		while (i < var_tupdesc->natts || j < tupdesc->natts)
+		while (attn1 < var_tupdesc->natts || attn2 < tupdesc->natts)
 		{
-			if (!target_field_is_valid && i < var_tupdesc->natts)
+			if (!target_field_is_valid && attn1 < var_tupdesc->natts)
 			{
-				tattr = TupleDescAttr(var_tupdesc, i);
+				tattr = TupleDescAttr(var_tupdesc, attn1);
 				if (tattr->attisdropped)
 				{
-					i += 1;
+					attn1 += 1;
 					continue;
 				}
 				target_field_is_valid = true;
 				target_nfields += 1;
 			}
 
-			if (!src_field_is_valid && j < tupdesc->natts)
+			if (!src_field_is_valid && attn2 < tupdesc->natts)
 			{
-				sattr = TupleDescAttr(tupdesc, j);
+				sattr = TupleDescAttr(tupdesc, attn2);
 				if (sattr->attisdropped)
 				{
-					j += 1;
+					attn2 += 1;
 					continue;
 				}
 				src_field_is_valid = true;
@@ -653,8 +663,8 @@ plpgsql_check_recval_assign_tupdesc(PLpgSQL_checkstate *cstate, PLpgSQL_rec *rec
 				/* try to search next tuple of fields */
 				src_field_is_valid =  false;
 				target_field_is_valid = false;
-				i += 1;
-				j += 1;
+				attn1 += 1;
+				attn2 += 1;
 			}
 			else
 				break;
@@ -767,6 +777,7 @@ compatible_tupdescs(TupleDesc src_tupdesc, TupleDesc dst_tupdesc)
 				return false;
 		}
 	}
+
 	return true;
 }
 
