@@ -26,18 +26,14 @@
 #include "catalog/pg_proc.h"
 #include "catalog/pg_type.h"
 #include "commands/extension.h"
+#include "commands/proclang.h"
 #include "utils/builtins.h"
 #include "utils/catcache.h"
 #include "utils/fmgroids.h"
 #include "utils/lsyscache.h"
 #include "utils/syscache.h"
-
-#if PG_VERSION_NUM >= 100000
-
 #include "utils/regproc.h"
 
-#endif
-
 #if PG_VERSION_NUM >= 110000
 
 #include "catalog/pg_proc.h"
@@ -50,9 +46,10 @@
 
 #endif
 
-
 #include "utils/syscache.h"
 
+static Oid plpgsql_check_PLpgSQLlanguageId = InvalidOid;
+
 /*
  * Fix - change of typename in Postgres 14
  */
@@ -158,25 +155,25 @@ void
 plpgsql_check_precheck_conditions(plpgsql_check_info *cinfo)
 {
 	Form_pg_proc proc;
-	Form_pg_language languageStruct;
-	HeapTuple	languageTuple;
 	char	   *funcname;
 
 	proc = (Form_pg_proc) GETSTRUCT(cinfo->proctuple);
 	funcname = format_procedure(cinfo->fn_oid);
 
-	/* used language must be plpgsql */
-	languageTuple = SearchSysCache1(LANGOID, ObjectIdGetDatum(proc->prolang));
-	Assert(HeapTupleIsValid(languageTuple));
+	/*
+	 * The plpgsql_check can be loaded by shared_proload_libraries. That means
+	 * so in init time the access to system catalog can be impossible. So
+	 * plpgsql_check_PLpgSQLlanguageId should be initialized here.
+	 */
+	if (!OidIsValid(plpgsql_check_PLpgSQLlanguageId))
+		plpgsql_check_PLpgSQLlanguageId = get_language_oid("plpgsql", false);
 
-	languageStruct = (Form_pg_language) GETSTRUCT(languageTuple);
-	if (strcmp(NameStr(languageStruct->lanname), "plpgsql") != 0)
+	/* used language must be plpgsql */
+	if (proc->prolang != plpgsql_check_PLpgSQLlanguageId)
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
 				 errmsg("%s is not a plpgsql function", funcname)));
 
-	ReleaseSysCache(languageTuple);
-
 	/* profiler doesn't require trigger data check */
 	if (!cinfo->show_profile)
 	{
@@ -312,3 +309,35 @@ plpgsql_check_pragma_func_oid(void)
 
 	return result;
 }
+
+/*
+ * Returns true, if function specified by oid is plpgsql function.
+ */
+bool
+plpgsql_check_is_plpgsql_function(Oid foid)
+{
+	HeapTuple	procTuple;
+	Form_pg_proc procStruct;
+	bool		result;
+
+	procTuple = SearchSysCache1(PROCOID, ObjectIdGetDatum(foid));
+	if (!HeapTupleIsValid(procTuple))
+		return false;
+
+	procStruct = (Form_pg_proc) GETSTRUCT(procTuple);
+
+	/*
+	 * The plpgsql_check can be loaded by shared_proload_libraries. That means
+	 * so in init time the access to system catalog can be impossible. So
+	 * plpgsql_check_PLpgSQLlanguageId should be initialized here.
+	 */
+	if (!OidIsValid(plpgsql_check_PLpgSQLlanguageId))
+		plpgsql_check_PLpgSQLlanguageId = get_language_oid("plpgsql", false);
+
+	result = procStruct->prolang == plpgsql_check_PLpgSQLlanguageId;
+
+	ReleaseSysCache(procTuple);
+
+	return result;
+}
+
