@@ -204,7 +204,7 @@ collect_volatility(PLpgSQL_checkstate *cstate, Query *query)
 CachedPlanSource *
 plpgsql_check_get_plan_source(PLpgSQL_checkstate *cstate, SPIPlanPtr plan)
 {
-	CachedPlanSource *plansource;
+	CachedPlanSource *plansource = NULL;
 
 	if (plan == NULL || plan->magic != _SPI_PLAN_MAGIC)
 		elog(ERROR, "cached plan is not valid plan");
@@ -225,7 +225,7 @@ plpgsql_check_get_plan_source(PLpgSQL_checkstate *cstate, SPIPlanPtr plan)
 			cstate->has_mp = true;
 		}
 		else
-			elog(ERROR, "plan is not single execution planyy");
+			elog(ERROR, "plan is not single execution plany");
 	}
 	else
 		plansource = (CachedPlanSource *) linitial(plan->plancache_list);
@@ -292,28 +292,52 @@ ExprGetQuery(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 	/* the test of PRAGMA function call */
 	if (result->commandType == CMD_SELECT)
 	{
-
-#if PG_VERSION_NUM < 100000
-
-		if (plansource->raw_parse_tree &&
-			IsA(plansource->raw_parse_tree, SelectStmt))
-		{
-			SelectStmt *selectStmt = (SelectStmt *) plansource->raw_parse_tree;
-
-#else
-
 		if (plansource->raw_parse_tree &&
 			plansource->raw_parse_tree->stmt &&
 			IsA(plansource->raw_parse_tree->stmt, SelectStmt))
 		{
 			SelectStmt *selectStmt = (SelectStmt *) plansource->raw_parse_tree->stmt;
 
-#endif
 			if (selectStmt->targetList && IsA(linitial(selectStmt->targetList), ResTarget))
 			{
 				ResTarget *rt = (ResTarget *) linitial(selectStmt->targetList);
 
-				if (rt->val && IsA(rt->val, FuncCall))
+				if (rt->val && IsA(rt->val, A_Const))
+				{
+					A_Const	   *ac = (A_Const *) rt->val;
+					bool		is_perform_stmt;
+					char *str = NULL;
+
+					is_perform_stmt = (cstate->estate &&
+									   cstate->estate->err_stmt &&
+									   cstate->estate->err_stmt->cmd_type == PLPGSQL_STMT_PERFORM);
+
+#if PG_VERSION_NUM < 150000
+
+					if (ac->val.type == T_String)
+						str = strVal(&(ac->val));
+
+#else
+
+					if (!ac->isnull && IsA(&ac->val, String))
+						str = strVal(&(ac->val));
+
+#endif
+
+					if (str && is_perform_stmt)
+					{
+						while (*str == ' ')
+							str++;
+
+						if (strncasecmp(str, "pragma:", 7) == 0)
+						{
+							cstate->was_pragma = true;
+
+							plpgsql_check_pragma_apply(cstate, str + 7, expr->ns, cstate->estate->err_stmt->lineno);
+						}
+					}
+				}
+				else if (rt->val && IsA(rt->val, FuncCall))
 				{
 					char	   *funcname;
 					char	   *schemaname;
@@ -335,8 +359,18 @@ ExprGetQuery(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 							{
 								A_Const *ac = (A_Const *) arg;
 
+#if PG_VERSION_NUM < 150000
+
 								if (ac->val.type == T_String)
-									plpgsql_check_pragma_apply(cstate, strVal(&(ac->val)));
+									plpgsql_check_pragma_apply(cstate, strVal(&(ac->val)), expr->ns, cstate->estate->err_stmt->lineno);
+
+#else
+
+								if (!ac->isnull && IsA(&ac->val, String))
+									plpgsql_check_pragma_apply(cstate, strVal(&(ac->val)), expr->ns, cstate->estate->err_stmt->lineno);
+
+#endif
+
 							}
 						}
 					}
@@ -371,13 +405,9 @@ get_cached_plan(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool *has_result
 
 	cplan = GetCachedPlan(plansource, NULL, NULL, NULL);
 
-#elif PG_VERSION_NUM >= 100000
-
-	cplan = GetCachedPlan(plansource, NULL, true, NULL);
-
 #else
 
-	cplan = GetCachedPlan(plansource, NULL, true);
+	cplan = GetCachedPlan(plansource, NULL, true, NULL);
 
 #endif
 
@@ -415,16 +445,8 @@ prohibit_write_plan(PLpgSQL_checkstate *cstate, CachedPlan *cplan, char *query_s
 	foreach(lc, cplan->stmt_list)
 	{
 
-#if PG_VERSION_NUM >= 100000
-
 		PlannedStmt *pstmt = (PlannedStmt *) lfirst(lc);
 
-#else
-
-		Node *pstmt = (Node *) lfirst(lc);
-
-#endif
-
 		if (!CommandIsReadOnly(pstmt))
 		{
 			StringInfoData message;
@@ -471,8 +493,6 @@ prohibit_transaction_stmt(PLpgSQL_checkstate *cstate, CachedPlan *cplan, char *q
 	{
 		Node *pstmt = (Node *) lfirst(lc);
 
-#if PG_VERSION_NUM >= 100000
-
 		/* PostgtreSQL 10 can have one level of nesting more */
 		if (IsA(pstmt, PlannedStmt))
 		{
@@ -482,8 +502,6 @@ prohibit_transaction_stmt(PLpgSQL_checkstate *cstate, CachedPlan *cplan, char *q
 				pstmt = (Node *) planstmt->utilityStmt;
 		}
 
-#endif
-
 		if (IsA(pstmt, TransactionStmt))
 		{
 			plpgsql_check_put_error(cstate,
@@ -749,7 +767,7 @@ plpgsql_check_expr_with_scalar_type(PLpgSQL_checkstate *cstate,
 
 		ReleaseTupleDesc(tupdesc);
 
-		RollbackAndReleaseCurrentSubTransaction();
+		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
 
@@ -865,7 +883,7 @@ plpgsql_check_returned_expr(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr, bool
 			ReleaseTupleDesc(tupdesc);
 		}
 
-		RollbackAndReleaseCurrentSubTransaction();
+		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
 
@@ -943,6 +961,25 @@ plpgsql_check_expr_as_rvalue(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr,
 
 #endif
 
+	/*
+	 * SELECT INTO for composite target type doesn't do
+	 * expand.
+	 */
+	if (targetrec || targetrow)
+	{
+		if (cstate->estate)
+		{
+			PLpgSQL_stmt *stmt = cstate->estate->err_stmt;
+
+			if (stmt &&
+				(stmt->cmd_type == PLPGSQL_STMT_EXECSQL ||
+				 stmt->cmd_type == PLPGSQL_STMT_DYNEXECUTE))
+			{
+				expand = false;
+			}
+		}
+	}
+
 	oldowner = CurrentResourceOwner;
 	BeginInternalSubTransaction(NULL);
 	MemoryContextSwitchTo(oldCxt);
@@ -1119,6 +1156,7 @@ plpgsql_check_expr_as_rvalue(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr,
 		{
 			if (targetrow != NULL || targetrec != NULL)
 				plpgsql_check_assign_tupdesc_row_or_rec(cstate, targetrow, targetrec, tupdesc, is_immutable_null);
+
 			if (targetdno != -1)
 				plpgsql_check_assign_tupdesc_dno(cstate, targetdno, tupdesc, is_immutable_null);
 
@@ -1147,7 +1185,7 @@ no_other_check:
 		if (tupdesc)
 			ReleaseTupleDesc(tupdesc);
 
-		RollbackAndReleaseCurrentSubTransaction();
+		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
 
@@ -1241,7 +1279,7 @@ plpgsql_check_expr_as_sqlstmt(PLpgSQL_checkstate *cstate, PLpgSQL_expr *expr)
 			ReleaseTupleDesc(tupdesc);
 		}
 
-		RollbackAndReleaseCurrentSubTransaction();
+		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
 
