@@ -890,12 +890,11 @@ plpgsql_check_setup_fcinfo(plpgsql_check_info *cinfo,
 
 	if (nargs > 0)
 	{
-		Oid		argtype = InvalidOid;
 		int		i;
 
 		for (i = 0; i < nargs; i++)
 		{
-			argtype = InvalidOid;
+			Oid		argtype = InvalidOid;
 
 			if (argmodes)
 			{
@@ -951,8 +950,7 @@ plpgsql_check_setup_fcinfo(plpgsql_check_info *cinfo,
 			for (i = 0; i < nargs; i++)
 			{
 				bool	is_variadic = false;
-
-				argtype = InvalidOid;
+				Oid		argtype = InvalidOid;
 
 				if (argmodes)
 				{
@@ -1176,8 +1174,6 @@ setup_estate(PLpgSQL_execstate *estate,
 
 #endif
 
-#if PG_VERSION_NUM > 100000
-
 	if (cinfo->oldtable)
 	{
 		EphemeralNamedRelation enr = palloc(sizeof(EphemeralNamedRelationData));
@@ -1210,12 +1206,6 @@ setup_estate(PLpgSQL_execstate *estate,
 		Assert(rc >= 0);
 	}
 
-#else
-
-		(void) cinfo;
-
-#endif
-
 	estate->err_stmt = NULL;
 	estate->err_text = NULL;
 
@@ -1284,6 +1274,7 @@ setup_cstate(PLpgSQL_checkstate *cstate,
 	cstate->safe_variables = NULL;
 	cstate->protected_variables = NULL;
 	cstate->auto_variables = NULL;
+	cstate->typed_variables = NULL;
 
 	cstate->stop_check = false;
 	cstate->allow_mp = false;
