@@ -21,12 +21,6 @@
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
 
-#if PG_VERSION_NUM < 90600
-
-#include "optimizer/planmain.h"
-
-#endif
-
 typedef struct 
 {
 	PLpgSQL_checkstate *cstate;
@@ -138,17 +132,6 @@ plpgsql_check_detect_dependency(PLpgSQL_checkstate *cstate, Query *query)
 	detect_dependency_walker((Node *) query, cstate);
 }
 
-/*
- * Expecting persistent oid of nextval, currval and setval functions.
- * Ensured by regress tests.
- */
-#define NEXTVAL_OID		1574
-#define CURRVAL_OID		1575
-#define SETVAL_OID		1576
-#define SETVAL2_OID		1765
-#define FORMAT_0PARAM_OID	3540
-#define FORMAT_NPARAM_OID	3539
-
 /*
  * When sequence related functions has constant oid parameter, then ensure, so
  * this oid is related to some sequnce object.
@@ -468,7 +451,7 @@ text_format_parse_format(const char *start_ptr,
 	else
 	{
 		/* Check for direct width specification */
-		found = text_format_parse_digits(&cp, end_ptr, &n, location, wp, is_error);
+		(void) text_format_parse_digits(&cp, end_ptr, &n, location, wp, is_error);
 		if (*is_error)
 			return NULL;
 	}
@@ -592,9 +575,10 @@ check_fmt_string(const char *fmt,
 				/* this is usually called after format check, but better be safe*/
 				if (argn <= nargs)
 				{
-					Node *arg = list_nth(args, argn - 1);
-
-					if (plpgsql_check_is_sql_injection_vulnerable(wp->cstate, wp->expr, arg, unsafe_expr_location))
+					if (plpgsql_check_is_sql_injection_vulnerable(wp->cstate,
+																  wp->expr,
+																  list_nth(args, argn - 1),
+																  unsafe_expr_location))
 					{
 						/* found vulnerability, stop */
 						*is_error = false;
@@ -938,9 +922,6 @@ plpgsql_check_is_sql_injection_vulnerable(PLpgSQL_checkstate *cstate,
 		return false;
 }
 
-
-#if PG_VERSION_NUM >= 90600
-
 /*
  * These checker function returns volatility or immutability of any function.
  * Special case is plpgsql_check_pragma function. This function is vollatile,
@@ -980,16 +961,12 @@ contain_volatile_functions_walker(Node *node, void *context)
 								context))
 		return true;
 
-#if PG_VERSION_NUM >= 100000
-
 	if (IsA(node, NextValueExpr))
 	{
 		/* NextValueExpr is volatile */
 		return true;
 	}
 
-#endif
-
 	/*
 	 * See notes in contain_mutable_functions_walker about why we treat
 	 * MinMaxExpr, XmlExpr, and CoerceToDomain as immutable, while
@@ -1019,8 +996,6 @@ contain_mutable_functions_walker(Node *node, void *context)
 								context))
 		return true;
 
-#if PG_VERSION_NUM >= 100000
-
 	if (IsA(node, SQLValueFunction))
 	{
 		/* all variants of SQLValueFunction are stable */
@@ -1033,8 +1008,6 @@ contain_mutable_functions_walker(Node *node, void *context)
 		return true;
 	}
 
-#endif
-
 	/*
 	 * It should be safe to treat MinMaxExpr as immutable, because it will
 	 * depend on a non-cross-type btree comparison function, and those should
@@ -1057,226 +1030,6 @@ contain_mutable_functions_walker(Node *node, void *context)
 								  context);
 }
 
-#else
-
-static bool
-contain_mutable_functions_walker(Node *node, void *context)
-{
-	if (node == NULL)
-		return false;
-	if (IsA(node, FuncExpr))
-	{
-		FuncExpr   *expr = (FuncExpr *) node;
-		PLpgSQL_checkstate *cstate = (PLpgSQL_checkstate *) context;
-
-		if (func_volatile(expr->funcid) != PROVOLATILE_IMMUTABLE &&
-				expr->funcid != cstate->pragma_foid)
-			return true;
-
-		/* else fall through to check args */
-	}
-	else if (IsA(node, OpExpr))
-	{
-		OpExpr	   *expr = (OpExpr *) node;
-
-		set_opfuncid(expr);
-		if (func_volatile(expr->opfuncid) != PROVOLATILE_IMMUTABLE)
-			return true;
-
-		/* else fall through to check args */
-	}
-	else if (IsA(node, DistinctExpr))
-	{
-		DistinctExpr *expr = (DistinctExpr *) node;
-
-		set_opfuncid((OpExpr *) expr);	/* rely on struct equivalence */
-		if (func_volatile(expr->opfuncid) != PROVOLATILE_IMMUTABLE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, NullIfExpr))
-	{
-		NullIfExpr *expr = (NullIfExpr *) node;
-
-		set_opfuncid((OpExpr *) expr);	/* rely on struct equivalence */
-		if (func_volatile(expr->opfuncid) != PROVOLATILE_IMMUTABLE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, ScalarArrayOpExpr))
-	{
-		ScalarArrayOpExpr *expr = (ScalarArrayOpExpr *) node;
-
-		set_sa_opfuncid(expr);
-		if (func_volatile(expr->opfuncid) != PROVOLATILE_IMMUTABLE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, CoerceViaIO))
-	{
-		CoerceViaIO *expr = (CoerceViaIO *) node;
-		Oid			iofunc;
-		Oid			typioparam;
-		bool		typisvarlena;
-
-		/* check the result type's input function */
-		getTypeInputInfo(expr->resulttype,
-						 &iofunc, &typioparam);
-		if (func_volatile(iofunc) != PROVOLATILE_IMMUTABLE)
-			return true;
-
-		/* check the input type's output function */
-		getTypeOutputInfo(exprType((Node *) expr->arg),
-						  &iofunc, &typisvarlena);
-		if (func_volatile(iofunc) != PROVOLATILE_IMMUTABLE)
-			return true;
-
-		/* else fall through to check args */
-	}
-	else if (IsA(node, ArrayCoerceExpr))
-	{
-		ArrayCoerceExpr *expr = (ArrayCoerceExpr *) node;
-
-		if (OidIsValid(expr->elemfuncid) &&
-			func_volatile(expr->elemfuncid) != PROVOLATILE_IMMUTABLE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, RowCompareExpr))
-	{
-		RowCompareExpr *rcexpr = (RowCompareExpr *) node;
-		ListCell   *opid;
-
-		foreach(opid, rcexpr->opnos)
-		{
-			if (op_volatile(lfirst_oid(opid)) != PROVOLATILE_IMMUTABLE)
-				return true;
-		}
-		/* else fall through to check args */
-	}
-	else if (IsA(node, Query))
-	{
-		/* Recurse into subselects */
-		return query_tree_walker((Query *) node,
-								 contain_mutable_functions_walker,
-								 context, 0);
-	}
-	return expression_tree_walker(node, contain_mutable_functions_walker,
-								  context);
-}
-
-/*
- * General purpose code for checking expression volatility.
- *
- * Special purpose code for use in COPY is almost identical to this,
- * so any changes here may also be needed in other contain_volatile...
- * functions.
- */
-static bool
-contain_volatile_functions_walker(Node *node, void *context)
-{
-	if (node == NULL)
-		return false;
-	if (IsA(node, FuncExpr))
-	{
-		FuncExpr   *expr = (FuncExpr *) node;
-		PLpgSQL_checkstate *cstate = (PLpgSQL_checkstate *) context;
-
-		if (func_volatile(expr->funcid) == PROVOLATILE_VOLATILE &&
-				expr->funcid != cstate->pragma_foid)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, OpExpr))
-	{
-		OpExpr	   *expr = (OpExpr *) node;
-
-		set_opfuncid(expr);
-		if (func_volatile(expr->opfuncid) == PROVOLATILE_VOLATILE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, DistinctExpr))
-	{
-		DistinctExpr *expr = (DistinctExpr *) node;
-
-		set_opfuncid((OpExpr *) expr);	/* rely on struct equivalence */
-		if (func_volatile(expr->opfuncid) == PROVOLATILE_VOLATILE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, NullIfExpr))
-	{
-		NullIfExpr *expr = (NullIfExpr *) node;
-
-		set_opfuncid((OpExpr *) expr);	/* rely on struct equivalence */
-		if (func_volatile(expr->opfuncid) == PROVOLATILE_VOLATILE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, ScalarArrayOpExpr))
-	{
-		ScalarArrayOpExpr *expr = (ScalarArrayOpExpr *) node;
-
-		set_sa_opfuncid(expr);
-		if (func_volatile(expr->opfuncid) == PROVOLATILE_VOLATILE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, CoerceViaIO))
-	{
-		CoerceViaIO *expr = (CoerceViaIO *) node;
-		Oid			iofunc;
-		Oid			typioparam;
-		bool		typisvarlena;
-
-		/* check the result type's input function */
-		getTypeInputInfo(expr->resulttype,
-						 &iofunc, &typioparam);
-		if (func_volatile(iofunc) == PROVOLATILE_VOLATILE)
-			return true;
-		/* check the input type's output function */
-		getTypeOutputInfo(exprType((Node *) expr->arg),
-						  &iofunc, &typisvarlena);
-		if (func_volatile(iofunc) == PROVOLATILE_VOLATILE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, ArrayCoerceExpr))
-	{
-		ArrayCoerceExpr *expr = (ArrayCoerceExpr *) node;
-
-		if (OidIsValid(expr->elemfuncid) &&
-			func_volatile(expr->elemfuncid) == PROVOLATILE_VOLATILE)
-			return true;
-		/* else fall through to check args */
-	}
-	else if (IsA(node, RowCompareExpr))
-	{
-		/* RowCompare probably can't have volatile ops, but check anyway */
-		RowCompareExpr *rcexpr = (RowCompareExpr *) node;
-		ListCell   *opid;
-
-		foreach(opid, rcexpr->opnos)
-		{
-			if (op_volatile(lfirst_oid(opid)) == PROVOLATILE_VOLATILE)
-				return true;
-		}
-		/* else fall through to check args */
-	}
-	else if (IsA(node, Query))
-	{
-		/* Recurse into subselects */
-		return query_tree_walker((Query *) node,
-								 contain_volatile_functions_walker,
-								 context, 0);
-	}
-	return expression_tree_walker(node, contain_volatile_functions_walker,
-								  context);
-}
-
-#endif
-
 /*
  * This is same like Postgres buildin function, but it ignores used
  * plpgsql_check pragma function
