@@ -37,7 +37,6 @@ PG_MODULE_MAGIC;
 
 PLpgSQL_plugin **plpgsql_check_plugin_var_ptr;
 
-
 static PLpgSQL_plugin plugin_funcs = { plpgsql_check_profiler_func_init,
 									   plpgsql_check_on_func_beg,
 									   plpgsql_check_profiler_func_end,
@@ -47,7 +46,6 @@ static PLpgSQL_plugin plugin_funcs = { plpgsql_check_profiler_func_init,
 									   NULL};
 
 
-
 static const struct config_enum_entry plpgsql_check_mode_options[] = {
 	{"disabled", PLPGSQL_CHECK_MODE_DISABLED, false},
 	{"by_function", PLPGSQL_CHECK_MODE_BY_FUNCTION, false},
@@ -81,8 +79,10 @@ void			_PG_fini(void);
 
 shmem_startup_hook_type prev_shmem_startup_hook = NULL;
 
+bool plpgsql_check_regress_test_mode;
+
 /*
- * Linkage to function in plpgsql module
+ * Links to function in plpgsql module
  */
 plpgsql_check__build_datatype_t plpgsql_check__build_datatype_p;
 plpgsql_check__compile_t plpgsql_check__compile_p;
@@ -90,6 +90,8 @@ plpgsql_check__parser_setup_t plpgsql_check__parser_setup_p;
 plpgsql_check__stmt_typename_t plpgsql_check__stmt_typename_p;
 plpgsql_check__exec_get_datum_type_t plpgsql_check__exec_get_datum_type_p;
 plpgsql_check__recognize_err_condition_t plpgsql_check__recognize_err_condition_p;
+plpgsql_check__ns_lookup_t plpgsql_check__ns_lookup_p;
+
 
 /*
  * load_external_function retursn PGFunctions - we need generic function, so
@@ -131,27 +133,29 @@ _PG_init(void)
 	plpgsql_check__stmt_typename_p = (plpgsql_check__stmt_typename_t)
 		LOAD_EXTERNAL_FUNCTION("$libdir/plpgsql", "plpgsql_stmt_typename");
 
-#if PG_VERSION_NUM >= 90600
-
 	AssertVariableIsOfType(&plpgsql_exec_get_datum_type, plpgsql_check__exec_get_datum_type_t);
 	plpgsql_check__exec_get_datum_type_p = (plpgsql_check__exec_get_datum_type_t)
 		LOAD_EXTERNAL_FUNCTION("$libdir/plpgsql", "plpgsql_exec_get_datum_type");
 
-#else
-
-	AssertVariableIsOfType(&exec_get_datum_type, plpgsql_check__exec_get_datum_type_t);
-	plpgsql_check__exec_get_datum_type_p = (plpgsql_check__exec_get_datum_type_t)
-		LOAD_EXTERNAL_FUNCTION("$libdir/plpgsql", "exec_get_datum_type");
-
-#endif
-
 	AssertVariableIsOfType(&plpgsql_recognize_err_condition, plpgsql_check__recognize_err_condition_t);
 	plpgsql_check__recognize_err_condition_p = (plpgsql_check__recognize_err_condition_t)
 		LOAD_EXTERNAL_FUNCTION("$libdir/plpgsql", "plpgsql_recognize_err_condition");
 
+	AssertVariableIsOfType(&plpgsql_ns_lookup, plpgsql_check__ns_lookup_t);
+	plpgsql_check__ns_lookup_p = (plpgsql_check__ns_lookup_t)
+		LOAD_EXTERNAL_FUNCTION("$libdir/plpgsql", "plpgsql_ns_lookup");
+
 	plpgsql_check_plugin_var_ptr = (PLpgSQL_plugin **) find_rendezvous_variable( "PLpgSQL_plugin");
 	*plpgsql_check_plugin_var_ptr = &plugin_funcs;
 
+	DefineCustomBoolVariable("plpgsql_check.regress_test_mode",
+					    "reduces volatile output",
+					    NULL,
+					    &plpgsql_check_regress_test_mode,
+					    false,
+					    PGC_USERSET, 0,
+					    NULL, NULL, NULL);
+
 	DefineCustomEnumVariable("plpgsql_check.mode",
 					    "choose a mode for enhanced checking",
 					    NULL,
@@ -288,17 +292,9 @@ _PG_init(void)
 
 		RequestAddinShmemSpace(plpgsql_check_shmem_size());
 
-#if PG_VERSION_NUM >= 90600
-
 		RequestNamedLWLockTranche("plpgsql_check profiler", 1);
 		RequestNamedLWLockTranche("plpgsql_check fstats", 1);
 
-#else
-
-		RequestAddinLWLocks(2);
-
-#endif
-
 		/*
 		 * Install hooks.
 		 */
@@ -306,6 +302,12 @@ _PG_init(void)
 		shmem_startup_hook = plpgsql_check_profiler_shmem_startup;
 	}
 
+	plpgsql_check_next_needs_fmgr_hook = needs_fmgr_hook;
+	plpgsql_check_next_fmgr_hook = fmgr_hook;
+
+	needs_fmgr_hook = plpgsql_check_needs_fmgr_hook;
+	fmgr_hook = plpgsql_check_fmgr_hook;
+
 	inited = true;
 }
 
@@ -319,4 +321,7 @@ _PG_fini(void)
 
 	/* Be more correct, and clean rendezvous variable */
 	*plpgsql_check_plugin_var_ptr = NULL;
+
+	needs_fmgr_hook = plpgsql_check_next_needs_fmgr_hook;
+	fmgr_hook = plpgsql_check_next_fmgr_hook;
 }
