@@ -15,7 +15,6 @@ typedef uint32 pc_queryid;
 #define NOQUERYID	(0)
 #endif
 
-
 enum
 {
 	PLPGSQL_CHECK_ERROR,
@@ -75,6 +74,7 @@ typedef struct plpgsql_check_result_info
 	int			format;						/* produced / expected format */
 	Tuplestorestate	*tuple_store;			/* target tuple store */
 	TupleDesc	tupdesc;					/* target tuple store tuple descriptor */
+	MemoryContext			query_ctx;		/* memory context for string operations */
 	StringInfo	sinfo;						/* buffer for multi line one value output formats */
 	bool		init_tag;					/* true, when init tag should be created */
 } plpgsql_check_result_info;
@@ -139,6 +139,7 @@ typedef struct PLpgSQL_checkstate
 	Bitmapset	   *out_variables;			/* what variables are used as OUT variables */
 	Bitmapset	   *protected_variables;	/* what variables should be assigned internal only */
 	Bitmapset	   *auto_variables;			/* variables initialized, used by runtime */
+	Bitmapset	   *typed_variables;		/* record variables with assigned type by pragma TYPE */
 	bool			stop_check;				/* true after error when fatal_errors option is active */
 	bool			allow_mp;				/* true, when multiple plans in plancache are allowed */
 	bool			has_mp;					/* true, when multiple plan was used */
@@ -187,10 +188,10 @@ extern void plpgsql_check_put_error(PLpgSQL_checkstate *cstate, int sqlerrcode,
 extern void plpgsql_check_put_error_edata(PLpgSQL_checkstate *cstate, ErrorData *edata);
 extern void plpgsql_check_put_dependency(plpgsql_check_result_info *ri, char *type, Oid oid, char *schema, char *name, char *params);
 extern void plpgsql_check_put_profile(plpgsql_check_result_info *ri, Datum queryids_array, int lineno, int stmt_lineno,
-	int cmds_on_row, int exec_count, int64 us_total, Datum max_time_array, Datum processed_rows_array, char *source_row);
+	int cmds_on_row, int64 exec_count, int64 exec_count_err, int64 us_total, Datum max_time_array, Datum processed_rows_array, char *source_row);
 extern void plpgsql_check_put_profile_statement(plpgsql_check_result_info *ri, pc_queryid queryid, int stmtid, int parent_stmtid, const char *parent_note, int block_num, int lineno,
-	int64 exec_stmts, double total_time, double max_time, int64 processed_rows, char *stmtname);
-extern void plpgsql_check_put_profiler_functions_all_tb(plpgsql_check_result_info *ri, Oid funcoid, int64 exec_count,
+	int64 exec_stmts, int64 exec_count_err, double total_time, double max_time, int64 processed_rows, char *stmtname);
+extern void plpgsql_check_put_profiler_functions_all_tb(plpgsql_check_result_info *ri, Oid funcoid, int64 exec_count, int64 exec_count_err,
 	double total_time, double avg_time, double stddev_time, double min_time, double max_time);
 
 /*
@@ -201,6 +202,7 @@ extern void plpgsql_check_get_function_info(HeapTuple procTuple, Oid *rettype, c
 extern void plpgsql_check_precheck_conditions(plpgsql_check_info *cinfo);
 extern char *plpgsql_check_get_src(HeapTuple procTuple);
 extern Oid plpgsql_check_pragma_func_oid(void);
+extern bool plpgsql_check_is_plpgsql_function(Oid foid);
 
 /*
  * functions from tablefunc.c
@@ -296,9 +298,11 @@ extern PLpgSQL_row * plpgsql_check_CallExprGetRowTarget(PLpgSQL_checkstate *csta
 #endif
 
 /*
- * functions from parse_name.c
+ * functions from parser.c
  */
 extern Oid plpgsql_check_parse_name_or_signature(char *name_or_signature);
+extern bool plpgsql_check_pragma_type(PLpgSQL_checkstate *cstate, const char *str, PLpgSQL_nsitem *ns, int lineno);
+extern bool plpgsql_check_pragma_table(PLpgSQL_checkstate *cstate, const char *str, int lineno);
 
 /*
  * functions from profiler.c
@@ -306,6 +310,9 @@ extern Oid plpgsql_check_parse_name_or_signature(char *name_or_signature);
 extern bool plpgsql_check_profiler;
 extern int plpgsql_check_profiler_max_shared_chunks;
 
+extern needs_fmgr_hook_type		plpgsql_check_next_needs_fmgr_hook;
+extern fmgr_hook_type			plpgsql_check_next_fmgr_hook;
+
 extern void plpgsql_check_profiler_shmem_startup(void);
 
 extern Size plpgsql_check_shmem_size(void);
@@ -325,6 +332,8 @@ extern void plpgsql_check_profiler_iterate_over_all_profiles(plpgsql_check_resul
 extern void plpgsql_check_init_trace_info(PLpgSQL_execstate *estate);
 extern bool plpgsql_check_get_trace_info(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt, PLpgSQL_execstate **outer_estate, int *frame_num, int *level, instr_time *start_time);
 
+extern bool plpgsql_check_needs_fmgr_hook(Oid fn_oid);
+extern void plpgsql_check_fmgr_hook(FmgrHookEventType event, FmgrInfo *flinfo, Datum *private);
 
 #if PG_VERSION_NUM >= 120000
 
@@ -347,6 +356,8 @@ extern int plpgsql_check_tracer_errlevel;
 extern PGErrorVerbosity plpgsql_check_tracer_verbosity;
 extern PGErrorVerbosity plpgsql_check_trace_assert_verbosity;
 
+extern bool plpgsql_check_regress_test_mode;
+
 extern void plpgsql_check_tracer_on_func_beg(PLpgSQL_execstate *estate, PLpgSQL_function *func);
 extern void plpgsql_check_tracer_on_func_end(PLpgSQL_execstate *estate, PLpgSQL_function *func);
 extern void plpgsql_check_tracer_on_stmt_beg(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt);
@@ -358,7 +369,7 @@ extern void plpgsql_check_set_stmt_group_number(PLpgSQL_stmt *stmt, int *group_n
 /*
  * variables from pragma.c
  */
-extern void plpgsql_check_pragma_apply(PLpgSQL_checkstate *cstate, char *pragma_str);
+extern void plpgsql_check_pragma_apply(PLpgSQL_checkstate *cstate, char *pragma_str, PLpgSQL_nsitem *ns, int lineno);
 
 extern plpgsql_check_pragma_vector plpgsql_check_runtime_pragma_vector;
 extern bool plpgsql_check_runtime_pragma_vector_changed;
@@ -378,7 +389,7 @@ extern PLpgSQL_plugin **plpgsql_check_plugin_var_ptr;
 #endif
 
 /*
- * Linkage to function in plpgsql module
+ * Links to function in plpgsql module
  */
 
 #ifdef PLPGSQL_BUILD_DATATYPE_4
@@ -403,6 +414,11 @@ extern plpgsql_check__exec_get_datum_type_t plpgsql_check__exec_get_datum_type_p
 typedef int (*plpgsql_check__recognize_err_condition_t) (const char *condname, bool allow_sqlstate);
 extern plpgsql_check__recognize_err_condition_t plpgsql_check__recognize_err_condition_p;
 
+typedef PLpgSQL_nsitem *(*plpgsql_check__ns_lookup_t) (PLpgSQL_nsitem *ns_cur, bool localmode,
+													  const char *name1, const char *name2, const char *name3,
+													  int *names_used);
+extern plpgsql_check__ns_lookup_t plpgsql_check__ns_lookup_p;
+
 #define NEVER_READ_VARIABLE_TEXT		"never read variable \"%s\""
 #define NEVER_READ_VARIABLE_TEXT_CHECK_LENGTH		19
 #define UNUSED_PARAMETER_TEXT			"unused parameter \"%s\""
@@ -420,6 +436,17 @@ extern plpgsql_check__recognize_err_condition_t plpgsql_check__recognize_err_con
 #define OUT_COMPOSITE_IS_NOT_SINGLE_TEXT	"composite OUT variable \"%s\" is not single argument"
 #define UNSAFE_EXECUTE					"the expression used by EXECUTE command is possibly sql injection vulnerable"
 
+/*
+ * Expecting persistent oid of nextval, currval and setval functions.
+ * Ensured by regress tests.
+ */
+#define NEXTVAL_OID		1574
+#define CURRVAL_OID		1575
+#define SETVAL_OID		1576
+#define SETVAL2_OID		1765
+#define FORMAT_0PARAM_OID	3540
+#define FORMAT_NPARAM_OID	3539
+
 #ifndef TupleDescAttr
 #define TupleDescAttr(tupdesc, i) ((tupdesc)->attrs[(i)])
 #endif
@@ -447,3 +474,10 @@ extern plpgsql_check__recognize_err_condition_t plpgsql_check__recognize_err_con
 #endif
 
 #define FUNCS_PER_USER		128 /* initial table size */
+
+#ifdef _MSC_VER
+
+#define strcasecmp _stricmp
+#define strncasecmp _strnicmp
+
+#endif
