@@ -76,12 +76,23 @@ typedef struct fstats
 	fstats_hashkey key;
 	slock_t		mutex;
 	uint64		exec_count;
+	uint64		exec_count_err;
 	uint64		total_time;
 	double		total_time_xx;
 	uint64		min_time;
 	uint64		max_time;
 } fstats;
 
+/*
+ * This is used as cache for types of expressions of USING clause
+ * (EXECUTE like statements).
+ */
+typedef struct
+{
+	int			nparams;
+	Oid			paramtypes[FLEXIBLE_ARRAY_MEMBER];
+} query_params;
+
 /*
  * Attention - the commands that can contains nestested commands
  * has attached own time and nested statements time too.
@@ -94,9 +105,11 @@ typedef struct profiler_stmt
 	uint64		us_total;
 	uint64		rows;
 	uint64		exec_count;
+	uint64		exec_count_err;
 	instr_time	start_time;
 	instr_time	total;
 	bool		has_queryid;
+	query_params *qparams;
 } profiler_stmt;
 
 typedef struct profiler_stmt_reduced
@@ -107,6 +120,7 @@ typedef struct profiler_stmt_reduced
 	uint64		us_total;
 	uint64		rows;
 	uint64		exec_count;
+	uint64		exec_count_err;
 	bool		has_queryid;
 } profiler_stmt_reduced;
 
@@ -186,7 +200,7 @@ typedef struct profiler_info
 
 	profiler_profile *profile;
 	profiler_stmt *stmts;
-	PLpgSQL_function *function;
+	PLpgSQL_function *func;
 	instr_time	start_time;
 
 	/* tracer part */
@@ -194,6 +208,7 @@ typedef struct profiler_info
 	int			frame_num;
 	int			level;
 	PLpgSQL_execstate *near_outer_estate;
+	PLpgSQL_execstate *estate;
 
 	bool		disable_tracer;
 
@@ -233,6 +248,41 @@ enum
 	COVERAGE_BRANCHES
 };
 
+typedef struct fmgr_hook_private
+{
+	bool		use_plpgsql;
+	Datum		next_private;
+} fmgr_hook_private;
+
+#define		NESTED_STMTS_STACK_SIZE		64
+
+typedef struct profiler_stack
+{
+	profiler_info *pinfo;
+	struct profiler_stack *prev_pinfo;
+
+	/*
+	 * The behaviour of estate->err_stmt is different across
+	 * pg releases, so the most easy way is hold own err_stmt
+	 * initilized in stmt_beg.
+	 */
+	PLpgSQL_stmt *err_stmt;
+
+	PLpgSQL_stmt *nested_stmts[NESTED_STMTS_STACK_SIZE];
+
+	/*
+	 * The commands executed under same protected block shares
+	 * eval_context. The eval_context is created for every
+	 * protected block, but exception handler uses parent
+	 * eval_context. The change of eval_context can signalize
+	 * an entry/leave to protected section. The nonempty
+	 * current error can signalize so we are inside an exception
+	 * handler.
+	 */
+	ExprContext *eval_econtext[NESTED_STMTS_STACK_SIZE];
+	int			nested_stmts_count;
+} profiler_stack;
+
 static HTAB *profiler_HashTable = NULL;
 static HTAB *shared_profiler_chunks_HashTable = NULL;
 static HTAB *profiler_chunks_HashTable = NULL;
@@ -245,7 +295,14 @@ static profiler_shared_state *profiler_ss = NULL;
 static MemoryContext profiler_mcxt = NULL;
 static MemoryContext profiler_queryid_mcxt = NULL;
 
+static profiler_stack *top_pinfo = NULL;
+static ExprContext *curr_eval_econtext = NULL;
+
 bool plpgsql_check_profiler = true;
+
+needs_fmgr_hook_type	plpgsql_check_next_needs_fmgr_hook = NULL;
+fmgr_hook_type			plpgsql_check_next_fmgr_hook = NULL;
+
 /*
  * should be enough for project of 300K PLpgSQL rows.
  * It should to take about 24.4MB of shared memory.
@@ -262,10 +319,19 @@ PG_FUNCTION_INFO_V1(plpgsql_profiler_install_fake_queryid_hook);
 PG_FUNCTION_INFO_V1(plpgsql_profiler_remove_fake_queryid_hook);
 
 static void update_persistent_profile(profiler_info *pinfo, PLpgSQL_function *func);
-static PLpgSQL_expr *profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic);
-static pc_queryid profiler_get_queryid(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt, bool *has_queryid);
+static PLpgSQL_expr *profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic, List **params);
+static pc_queryid profiler_get_queryid(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt, bool *has_queryid, query_params **qparams);
+
+#if PG_VERSION_NUM >= 140000
+
+static void profiler_fake_queryid_hook(ParseState *pstate, Query *query, JumbleState *jstate);
+
+#else
+
 static void profiler_fake_queryid_hook(ParseState *pstate, Query *query);
 
+#endif
+
 static void profile_register_stmt(profiler_info *pinfo, profiler_stmt_walker_options *opts, PLpgSQL_stmt *stmt);
 static void stmts_walker(profiler_info *pinfo, profiler_stmt_walker_mode, List *stmts, PLpgSQL_stmt *parent_stmt,
 	 const char *description, profiler_stmt_walker_options *opts);
@@ -568,19 +634,8 @@ plpgsql_check_profiler_shmem_startup(void)
 
 	if (!found)
 	{
-
-#if PG_VERSION_NUM > 90600
-
 		profiler_ss->lock = &(GetNamedLWLockTranche("plpgsql_check profiler"))->lock;
 		profiler_ss->fstats_lock = &(GetNamedLWLockTranche("plpgsql_check fstats"))->lock;
-
-#else
-
-		profiler_ss->lock = LWLockAssign();
-		profiler_ss->fstats_lock = LWLockAssign();
-
-#endif
-
 	}
 
 	memset(&info, 0, sizeof(info));
@@ -734,21 +789,21 @@ increment_branch_counter(coverage_state *cs, int64 executed)
  */
 #if PG_VERSION_NUM < 120000
 
-#define STMTID(stmt)		(profiler_get_stmtid(pinfo->profile, pinfo->function, ((PLpgSQL_stmt *) stmt)))
+#define STMTID(stmt)		(profiler_get_stmtid(pinfo->profile, pinfo->func, ((PLpgSQL_stmt *) stmt)))
 #define FUNC_NSTATEMENTS(pinfo)	(pinfo->profile->nstatements)
 #define NATURAL_STMTID(pinfo, id) (id)
 
 #else
 
 #define STMTID(stmt)		(((PLpgSQL_stmt *) stmt)->stmtid - 1)
-#define FUNC_NSTATEMENTS(pinfo) ((int) pinfo->function->nstatements)
+#define FUNC_NSTATEMENTS(pinfo) ((int) pinfo->func->nstatements)
 
 static int
 get_natural_stmtid(profiler_info *pinfo, int id)
 {
 	int		i;
 
-	for (i = 0; i < ((int) pinfo->function->nstatements); i++)
+	for (i = 0; i < FUNC_NSTATEMENTS(pinfo); i++)
 		if (pinfo->profile->stmtid_reorder_map[i] == id)
 			return i;
 
@@ -844,10 +899,10 @@ profiler_stmt_walker(profiler_info *pinfo,
 	profiler_stmt *pstmt = NULL;
 	profiler_profile *profile = pinfo->profile;
 
-	bool	prepare_profile		= mode == PLPGSQL_CHECK_STMT_WALKER_PREPARE_PROFILE;
-	bool	count_exec_time		= mode == PLPGSQL_CHECK_STMT_WALKER_COUNT_EXEC_TIME;
-	bool	prepare_result		= mode == PLPGSQL_CHECK_STMT_WALKER_PREPARE_RESULT;
-	bool	collect_coverage	= mode == PLPGSQL_CHECK_STMT_WALKER_COLLECT_COVERAGE;
+	bool	prepare_profile_mode  = mode == PLPGSQL_CHECK_STMT_WALKER_PREPARE_PROFILE;
+	bool	count_exec_time_mode  = mode == PLPGSQL_CHECK_STMT_WALKER_COUNT_EXEC_TIME;
+	bool	prepare_result_mode	  = mode == PLPGSQL_CHECK_STMT_WALKER_PREPARE_RESULT;
+	bool	collect_coverage_mode = mode == PLPGSQL_CHECK_STMT_WALKER_COLLECT_COVERAGE;
 
 	int64	total_us_time = 0;
 	int64	nested_us_time = 0;
@@ -862,7 +917,7 @@ profiler_stmt_walker(profiler_info *pinfo,
 
 	Assert(profile);
 
-	if (prepare_profile)
+	if (prepare_profile_mode)
 	{
 		profile_register_stmt(pinfo, opts, stmt);
 	}
@@ -870,7 +925,7 @@ profiler_stmt_walker(profiler_info *pinfo,
 	{
 		stmtid = STMTID(stmt);
 
-		if (count_exec_time)
+		if (count_exec_time_mode)
 		{
 			/*
 			 * Get statement info from function execution context
@@ -912,7 +967,7 @@ profiler_stmt_walker(profiler_info *pinfo,
 			 */
 			ppstmt = get_stmt_profile_next(opts->pi);
 
-			if (prepare_result)
+			if (prepare_result_mode)
 			{
 				int parent_stmtid = parent_stmt ? ((int) STMTID(parent_stmt)) : -1;
 
@@ -926,13 +981,14 @@ profiler_stmt_walker(profiler_info *pinfo,
 														stmt_block_num,
 														stmt->lineno,
 														ppstmt ? ppstmt->exec_count : 0,
+														ppstmt ? ppstmt->exec_count_err : 0,
 														ppstmt ? ppstmt->us_total : 0.0,
 														ppstmt ? ppstmt->us_max : 0.0,
 														ppstmt ? ppstmt->rows : 0,
 														(char *) plpgsql_check__stmt_typename_p(stmt));
 				}
 			}
-			else if (collect_coverage)
+			else if (collect_coverage_mode)
 			{
 				/* save statement exec count */
 				exec_count = ppstmt ? ppstmt->exec_count : 0;
@@ -955,7 +1011,7 @@ profiler_stmt_walker(profiler_info *pinfo,
 					 stmts, stmt, "loop body",
 					 opts);
 
-		if (collect_coverage)
+		if (collect_coverage_mode)
 			increment_branch_counter(opts->cs,
 									 opts->nested_exec_count);
 	}
@@ -976,11 +1032,11 @@ profiler_stmt_walker(profiler_info *pinfo,
 					 stmt_if->then_body, stmt, "then body",
 					 opts);
 
-		if (count_exec_time)
+		if (count_exec_time_mode)
 		{
 			nested_us_time = opts->nested_us_time;
 		}
-		else if (collect_coverage)
+		else if (collect_coverage_mode)
 		{
 			increment_branch_counter(opts->cs,
 									 opts->nested_exec_count);
@@ -998,10 +1054,10 @@ profiler_stmt_walker(profiler_info *pinfo,
 						 stmts, stmt, strbuf,
 						 opts);
 
-			if (count_exec_time)
+			if (count_exec_time_mode)
 				nested_us_time += opts->nested_us_time;
 
-			else if (collect_coverage)
+			else if (collect_coverage_mode)
 			{
 				increment_branch_counter(opts->cs,
 										 opts->nested_exec_count);
@@ -1016,17 +1072,17 @@ profiler_stmt_walker(profiler_info *pinfo,
 						 stmt_if->else_body, stmt, "else body",
 						 opts);
 
-			if (count_exec_time)
+			if (count_exec_time_mode)
 				nested_us_time += opts->nested_us_time;
 
-			else if (collect_coverage)
+			else if (collect_coverage_mode)
 				increment_branch_counter(opts->cs,
 										 opts->nested_exec_count);
 		}
 		else
 		{
 			/* calculate exec_count for implicit else path */
-			if (collect_coverage)
+			if (collect_coverage_mode)
 			{
 				int64 else_exec_count = exec_count - all_nested_branches_exec_count;
 
@@ -1049,10 +1105,10 @@ profiler_stmt_walker(profiler_info *pinfo,
 						 stmts, stmt, strbuf,
 						 opts);
 
-			if (count_exec_time)
+			if (count_exec_time_mode)
 				nested_us_time = opts->nested_us_time;
 
-			else if (collect_coverage)
+			else if (collect_coverage_mode)
 				increment_branch_counter(opts->cs,
 										 opts->nested_exec_count);
 		}
@@ -1061,10 +1117,10 @@ profiler_stmt_walker(profiler_info *pinfo,
 					 stmt_case->else_stmts, stmt, "case else",
 					 opts);
 
-		if (count_exec_time)
+		if (count_exec_time_mode)
 			nested_us_time = opts->nested_us_time;
 
-		else if (collect_coverage)
+		else if (collect_coverage_mode)
 			increment_branch_counter(opts->cs,
 									 opts->nested_exec_count);
 	}
@@ -1076,11 +1132,15 @@ profiler_stmt_walker(profiler_info *pinfo,
 					 stmt_block->body, stmt, "body",
 					 opts);
 
-		if (count_exec_time)
+		if (count_exec_time_mode)
 			nested_us_time = opts->nested_us_time;
 
 		if (stmt_block->exceptions)
 		{
+			if (collect_coverage_mode)
+				increment_branch_counter(opts->cs,
+										 opts->nested_exec_count);
+
 			foreach(lc, stmt_block->exceptions->exc_list)
 			{
 				stmts = ((PLpgSQL_exception *) lfirst(lc))->action;
@@ -1091,13 +1151,16 @@ profiler_stmt_walker(profiler_info *pinfo,
 							 stmts, stmt, strbuf,
 							 opts);
 
-				if (count_exec_time)
+				if (count_exec_time_mode)
 					nested_us_time += opts->nested_us_time;
+				else if (collect_coverage_mode)
+					increment_branch_counter(opts->cs,
+											 opts->nested_exec_count);
 			}
 		}
 	}
 
-	if (count_exec_time)
+	if (count_exec_time_mode)
 	{
 		Assert (pstmt);
 
@@ -1111,7 +1174,7 @@ profiler_stmt_walker(profiler_info *pinfo,
 		if (pstmt->exec_count == 1 && pstmt->us_max == 1)
 			pstmt->us_max = pstmt->us_total;
 	}
-	else if (collect_coverage)
+	else if (collect_coverage_mode)
 	{
 		opts->nested_exec_count = exec_count;
 	}
@@ -1162,8 +1225,8 @@ plpgsql_profiler_reset_all(PG_FUNCTION_ARGS)
 
 		LWLockRelease(profiler_ss->fstats_lock);
 	}
-	else
-		plpgsql_check_profiler_init_hash_tables();
+
+	plpgsql_check_profiler_init_hash_tables();
 
 	PG_RETURN_VOID();
 }
@@ -1432,6 +1495,7 @@ update_persistent_fstats(PLpgSQL_function *func,
 	if (!found)
 	{
 		fstats_item->exec_count = 0;
+		fstats_item->exec_count_err = 0;
 		fstats_item->total_time = 0;
 		fstats_item->total_time_xx = 0.0;
 		fstats_item->min_time = elapsed;
@@ -1572,6 +1636,7 @@ update_persistent_profile(profiler_info *pinfo,
 			prstmt->us_total = pstmt->us_total;
 			prstmt->rows = pstmt->rows;
 			prstmt->exec_count = pstmt->exec_count;
+			prstmt->exec_count_err = pstmt->exec_count_err;
 		}
 
 		/* clean unused stmts in chunk */
@@ -1670,6 +1735,7 @@ update_persistent_profile(profiler_info *pinfo,
 			prstmt->us_total += pstmt->us_total;
 			prstmt->rows += pstmt->rows;
 			prstmt->exec_count += pstmt->exec_count;
+			prstmt->exec_count_err += pstmt->exec_count_err;
 		}
 	}
 	PG_CATCH();
@@ -1769,7 +1835,7 @@ profile_register_stmt(profiler_info *pinfo,
 {
 #if PG_VERSION_NUM < 120000
 
-		profiler_update_map(pinfo->profile, opts, pinfo->function, stmt);
+		profiler_update_map(pinfo->profile, opts, pinfo->func, stmt);
 
 #else
 
@@ -1893,10 +1959,11 @@ stmts_walker(profiler_info *pinfo,
  * queryid.
  */
 static PLpgSQL_expr  *
-profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic)
+profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic, List **params)
 {
 	PLpgSQL_expr *expr = NULL;
 
+	*params = NIL;
 	*dynamic = false;
 
 	switch(stmt->cmd_type)
@@ -1913,11 +1980,13 @@ profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic)
 		case PLPGSQL_STMT_CALL:
 			expr = ((PLpgSQL_stmt_call *) stmt)->expr;
 			break;
+#if PG_VERSION_NUM < 140000
 		case PLPGSQL_STMT_SET:
 			expr = ((PLpgSQL_stmt_set *) stmt)->expr;
 			break;
+#endif			/* PG_VERSION_NUM < 140000 */
 
-#endif
+#endif			/* PG_VERSION_NUM >= 110000 */
 
 		case PLPGSQL_STMT_IF:
 			expr = ((PLpgSQL_stmt_if *) stmt)->cond;
@@ -1933,6 +2002,8 @@ profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic)
 			break;
 		case PLPGSQL_STMT_DYNFORS:
 			expr = ((PLpgSQL_stmt_dynfors *) stmt)->query;
+			*params = ((PLpgSQL_stmt_dynfors *) stmt)->params;
+			*dynamic = true;
 			break;
 		case PLPGSQL_STMT_FOREACH_A:
 			expr = ((PLpgSQL_stmt_foreach_a *) stmt)->expr;
@@ -1959,6 +2030,7 @@ profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic)
 				else
 				{
 					expr = q->dynquery;
+					*params = q->params;
 					*dynamic = true;
 				}
 			}
@@ -1971,6 +2043,7 @@ profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic)
 			break;
 		case PLPGSQL_STMT_DYNEXECUTE:
 			expr = ((PLpgSQL_stmt_dynexecute *) stmt)->query;
+			*params = ((PLpgSQL_stmt_dynexecute *) stmt)->params;
 			*dynamic = true;
 			break;
 		case PLPGSQL_STMT_OPEN:
@@ -1989,13 +2062,8 @@ profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic)
 					expr = o->argquery;
 			}
 		case PLPGSQL_STMT_BLOCK:
-
-#if PG_VERSION_NUM >= 100000
-
 		case PLPGSQL_STMT_FORS:
 
-#endif
-
 #if PG_VERSION_NUM >= 110000
 
 		case PLPGSQL_STMT_COMMIT:
@@ -2015,26 +2083,24 @@ profiler_get_expr(PLpgSQL_stmt *stmt, bool *dynamic)
 }
 
 static pc_queryid
-profiler_get_dyn_queryid(PLpgSQL_execstate *estate, PLpgSQL_expr *expr)
+profiler_get_dyn_queryid(PLpgSQL_execstate *estate, PLpgSQL_expr *expr, query_params *qparams)
 {
 	MemoryContext oldcxt;
 	Query	   *query;
-
-#if PG_VERSION_NUM >= 100000
-
 	RawStmt    *parsetree;
-
-#else
-
-	Node	   *parsetree;
-
-#endif
-
 	bool		snapshot_set;
 	List	   *parsetree_list;
 	PLpgSQL_var result;
 	PLpgSQL_type typ;
 	char	   *query_string = NULL;
+	Oid		   *paramtypes = NULL;
+	int			nparams = 0;
+
+	if (qparams)
+	{
+		paramtypes = qparams->paramtypes;
+		nparams = qparams->nparams;
+	}
 
 	memset(&result, 0, sizeof(result));
 	memset(&typ, 0, sizeof(typ));
@@ -2082,16 +2148,7 @@ profiler_get_dyn_queryid(PLpgSQL_execstate *estate, PLpgSQL_expr *expr)
 	}
 
 	/* Run through the raw parsetree and process it. */
-#if PG_VERSION_NUM >= 100000
-
 	parsetree = (RawStmt *) linitial(parsetree_list);
-
-#else
-
-	parsetree = (Node *) linitial(parsetree_list);
-
-#endif
-
 	snapshot_set = false;
 
 	/*
@@ -2103,15 +2160,7 @@ profiler_get_dyn_queryid(PLpgSQL_execstate *estate, PLpgSQL_expr *expr)
 		snapshot_set = true;
 	}
 
-#if PG_VERSION_NUM >= 100000
-
-	query = parse_analyze(parsetree, query_string, NULL, 0, NULL);
-
-#else 
-
-	query = parse_analyze(parsetree, query_string, NULL, 0);
-
-#endif
+	query = parse_analyze(parsetree, query_string, paramtypes, nparams, NULL);
 
 	if (snapshot_set)
 		PopActiveSnapshot();
@@ -2122,44 +2171,105 @@ profiler_get_dyn_queryid(PLpgSQL_execstate *estate, PLpgSQL_expr *expr)
 	return query->queryId;
 }
 
+/*
+ * Returns result type of already executed (has assigned plan) expression
+ */
+static bool
+get_expr_type(PLpgSQL_expr *expr, Oid *result_type)
+{
+	if (expr)
+	{
+		SPIPlanPtr ptr  = expr->plan;
+
+		if (ptr)
+		{
+			List	   *plan_sources = SPI_plan_get_plan_sources(ptr);
+
+			if (plan_sources && list_length(plan_sources) == 1)
+			{
+				CachedPlanSource *plan_source;
+				TupleDesc	tupdesc;
+
+				plan_source = (CachedPlanSource *) linitial(plan_sources);
+				tupdesc = plan_source->resultDesc;
+
+				if (tupdesc->natts == 1)
+				{
+					*result_type = TupleDescAttr(tupdesc, 0)->atttypid;
+
+					return true;
+				}
+			}
+		}
+	}
+
+	return false;
+}
+
+
 /* Return the first queryid found in the given PLpgSQL_stmt, if any. */
 static pc_queryid
 profiler_get_queryid(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt,
-					 bool *has_queryid)
+					 bool *has_queryid, query_params **qparams)
 {
-	bool dynamic;
-	PLpgSQL_expr *expr = profiler_get_expr(stmt, &dynamic);
+	PLpgSQL_expr *expr;
+	bool		dynamic;
+	List	   *params;
+	List	   *plan_sources;
 
+	expr = profiler_get_expr(stmt, &dynamic, &params);
 	*has_queryid = (expr != NULL);
 
+	/* fast leaving, when expression has not assigned plan */
+	if (!expr || !expr->plan)
+		return NOQUERYID;
+
 	if (dynamic)
 	{
 		Assert(expr);
 
-		return profiler_get_dyn_queryid(estate, expr);
+		if (params && !*qparams)
+		{
+			query_params *qps = NULL;
+			int		nparams = list_length(params);
+			int		paramno = 0;
+			MemoryContext oldcxt;
+			ListCell *lc;
+
+			/* build array of Oid used like dynamic query parameters */
+			oldcxt = MemoryContextSwitchTo(profiler_mcxt);
+			qps = (query_params *) palloc(sizeof(Oid) * nparams + sizeof(int));
+			MemoryContextSwitchTo(oldcxt);
+
+			foreach(lc, params)
+			{
+				PLpgSQL_expr *param_expr = (PLpgSQL_expr *) lfirst(lc);
+
+				if (!get_expr_type(param_expr, &qps->paramtypes[paramno++]))
+				{
+					free(qps);
+					return NOQUERYID;
+				}
+			}
+
+			qps->nparams = nparams;
+			*qparams = qps;
+		}
+
+		return profiler_get_dyn_queryid(estate, expr, *qparams);
 	}
 
-	if (expr)
+	plan_sources = SPI_plan_get_plan_sources(expr->plan);
+
+	if (plan_sources)
 	{
-		SPIPlanPtr ptr  = expr->plan;
-		List *sources;
+		CachedPlanSource *plan_source = (CachedPlanSource *) linitial(plan_sources);
 
-		if (!ptr)
-			return NOQUERYID;
-
-		sources = SPI_plan_get_plan_sources(ptr);
-
-		if (sources)
+		if (plan_source->query_list)
 		{
-			CachedPlanSource *source;
+			Query *q = linitial_node(Query, plan_source->query_list);
 
-			source = (CachedPlanSource *) linitial(sources);
-			if (source->query_list)
-			{
-				Query *q = linitial_node(Query, source->query_list);
-
-				return q->queryId;
-			}
+			return q->queryId;
 		}
 	}
 
@@ -2170,9 +2280,21 @@ profiler_get_queryid(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt,
  * Generate simple queryid  for testing purpose.
  * DO NOT USE IN PRODUCTION.
  */
+#if PG_VERSION_NUM >= 140000
+
+static void
+profiler_fake_queryid_hook(ParseState *pstate, Query *query, JumbleState *jstate)
+{
+	(void) jstate;
+
+#else
+
 static void
 profiler_fake_queryid_hook(ParseState *pstate, Query *query)
 {
+
+#endif
+
 	(void) pstate;
 
 	Assert(query->queryId == NOQUERYID);
@@ -2200,7 +2322,7 @@ prepare_profile(profiler_info *pinfo,
 				bool init)
 {
 	profiler_stmt_walker_options opts;
-	PLpgSQL_function *function;
+	PLpgSQL_function *func;
 	int		i;
 
 #if PG_VERSION_NUM < 120000
@@ -2209,11 +2331,11 @@ prepare_profile(profiler_info *pinfo,
 
 #endif
 
-	Assert(pinfo && pinfo->function);
+	Assert(pinfo && pinfo->func);
 
 	memset(&opts, 0, sizeof(profiler_stmt_walker_options));
 
-	function = pinfo->function;
+	func = pinfo->func;
 	pinfo->profile = profile;
 
 	if (init)
@@ -2239,21 +2361,21 @@ prepare_profile(profiler_info *pinfo,
 
 		oldcxt = MemoryContextSwitchTo(profiler_mcxt);
 
-		profile->stmtid_reorder_map = palloc0(sizeof(int) * function->nstatements);
+		profile->stmtid_reorder_map = palloc0(sizeof(int) * func->nstatements);
 
 		/*
 		 * I found my bug in PLpgSQL runtime - when function statement counter
 		 * is incremental 2x for every FOR cycle. Until fix this bug, some entries
 		 * of reorder map can be uniinitialized, and we have to detect these entries.
 		 */
-		for (i = 0; i < ((int) function->nstatements); i++)
+		for (i = 0; i < ((int) func->nstatements); i++)
 			profile->stmtid_reorder_map[i] = -1;
 
 		MemoryContextSwitchTo(oldcxt);
 
 		opts.stmtid = 0;
 		profiler_stmt_walker(pinfo, PLPGSQL_CHECK_STMT_WALKER_PREPARE_PROFILE,
-							 (PLpgSQL_stmt *) function->action, NULL, NULL, 1,
+							 (PLpgSQL_stmt *) func->action, NULL, NULL, 1,
 							 &opts);
 
 #endif
@@ -2267,7 +2389,7 @@ prepare_profile(profiler_info *pinfo,
 	 */
 	for (i = 0; i < profile->n_mapped_functions; i++)
 	{
-		if (profile->mapped_functions[i] == function)
+		if (profile->mapped_functions[i] == func)
 		{
 			found = true;
 			break;
@@ -2293,11 +2415,11 @@ prepare_profile(profiler_info *pinfo,
 			profile->max_mapped_functions = new_max_mapped_functions;
 		}
 
-		profile->mapped_functions[profile->n_mapped_functions++] = function;
+		profile->mapped_functions[profile->n_mapped_functions++] = func;
 
 		opts.stmtid = 0;
 		profiler_stmt_walker(pinfo, PLPGSQL_CHECK_STMT_WALKER_PREPARE_PROFILE,
-							 (PLpgSQL_stmt *) function->action, NULL, NULL, 1,
+							 (PLpgSQL_stmt *) func->action, NULL, NULL, 1,
 							 &opts);
 
 		if (profile->nstatements > 0 && profile->nstatements != opts.stmtid)
@@ -2397,9 +2519,9 @@ plpgsql_check_iterate_over_profile(plpgsql_check_info *cinfo,
 								   &tg_trigger,
 								   &fake_rtd);
 
-		pinfo.function = plpgsql_check__compile_p(fake_fcinfo, false);
+		pinfo.func = plpgsql_check__compile_p(fake_fcinfo, false);
 
-		profiler_init_hashkey(&hk_function, pinfo.function);
+		profiler_init_hashkey(&hk_function, pinfo.func);
 		profile = (profiler_profile *) hash_search(profiler_HashTable,
 												 (void *) &hk_function,
 												 HASH_ENTER,
@@ -2410,7 +2532,7 @@ plpgsql_check_iterate_over_profile(plpgsql_check_info *cinfo,
 		opts.pi =  &pi;
 		opts.cs = cs;
 
-		stmt = (PLpgSQL_stmt *) pinfo.function->action;
+		stmt = (PLpgSQL_stmt *) pinfo.func->action;
 
 		profiler_stmt_walker(&pinfo, mode, stmt, NULL, NULL, 1, &opts);
 	}
@@ -2494,6 +2616,7 @@ plpgsql_check_profiler_show_profile(plpgsql_check_result_info *ri,
 			int			stmt_lineno = -1;
 			int64		us_total = 0;
 			int64		exec_count = 0;
+			int64		exec_count_err = 0;
 			Datum		queryids_array = (Datum) 0;
 			Datum		max_time_array = (Datum) 0;
 			Datum		processed_rows_array = (Datum) 0;
@@ -2521,14 +2644,10 @@ plpgsql_check_profiler_show_profile(plpgsql_check_result_info *ri,
 				ArrayBuildState *max_time_abs = NULL;
 				ArrayBuildState *processed_rows_abs = NULL;
 
-#if PG_VERSION_NUM >= 90500
-
 				queryids_abs = initArrayResult(INT8OID, CurrentMemoryContext, true);
 				max_time_abs = initArrayResult(FLOAT8OID, CurrentMemoryContext, true);
 				processed_rows_abs = initArrayResult(INT8OID, CurrentMemoryContext, true);
 
-#endif
-
 				/* process all statements on this line */
 				for(;;)
 				{
@@ -2565,6 +2684,7 @@ plpgsql_check_profiler_show_profile(plpgsql_check_result_info *ri,
 
 						us_total += prstmt->us_total;
 						exec_count += prstmt->exec_count;
+						exec_count_err += prstmt->exec_count_err;
 
 						stmt_lineno = lineno;
 
@@ -2614,6 +2734,7 @@ plpgsql_check_profiler_show_profile(plpgsql_check_result_info *ri,
 								   stmt_lineno,
 								   cmds_on_row,
 								   exec_count,
+								   exec_count_err,
 								   us_total,
 								   max_time_array,
 								   processed_rows_array,
@@ -2647,7 +2768,7 @@ init_profiler_info(profiler_info *pinfo, PLpgSQL_function *func)
 	{
 		pinfo = palloc0(sizeof(profiler_info));
 		pinfo->pi_magic = PI_MAGIC;
-		pinfo->function = func;
+		pinfo->func = func;
 	}
 
 	return pinfo;
@@ -2718,19 +2839,35 @@ plpgsql_check_profiler_func_init(PLpgSQL_execstate *estate, PLpgSQL_function *fu
 	}
 
 	if (pinfo)
+	{
 		INSTR_TIME_SET_CURRENT(pinfo->start_time);
 
+		pinfo->estate = estate;
+	}
+
 	estate->plugin_info = pinfo;
+
+	if (top_pinfo)
+	{
+		top_pinfo->pinfo = pinfo;
+		curr_eval_econtext = estate->eval_econtext;
+	}
 }
 
 void
 plpgsql_check_profiler_func_end(PLpgSQL_execstate *estate, PLpgSQL_function *func)
 {
-	profiler_info *pinfo = estate->plugin_info;
+	profiler_info *pinfo = NULL;
+
+	if (estate)
+		pinfo = (profiler_info *) estate->plugin_info;
+	else if (top_pinfo)
+		pinfo = top_pinfo->pinfo;
 
 	if (plpgsql_check_tracer && pinfo )
 	{
-		plpgsql_check_tracer_on_func_end(estate, func);
+		if (estate)
+			plpgsql_check_tracer_on_func_end(estate, func);
 
 #if PG_VERSION_NUM >= 120000
 
@@ -2748,8 +2885,7 @@ plpgsql_check_profiler_func_end(PLpgSQL_execstate *estate, PLpgSQL_function *fun
 		pinfo && pinfo->profile &&
 		func->fn_oid != InvalidOid)
 	{
-		profiler_info *pinfo = (profiler_info *) estate->plugin_info;
-		int		entry_stmtid = STMTID(pinfo->function->action);
+		int		entry_stmtid = STMTID(pinfo->func->action);
 
 		instr_time		end_time;
 		uint64			elapsed;
@@ -2765,6 +2901,7 @@ plpgsql_check_profiler_func_end(PLpgSQL_execstate *estate, PLpgSQL_function *fun
 		if (pinfo->stmts[entry_stmtid].exec_count == 0)
 		{
 			pinfo->stmts[entry_stmtid].exec_count = 1;
+			pinfo->stmts[entry_stmtid].exec_count_err = 0;
 			pinfo->stmts[entry_stmtid].us_total = elapsed;
 			pinfo->stmts[entry_stmtid].us_max = elapsed;
 		}
@@ -2772,7 +2909,7 @@ plpgsql_check_profiler_func_end(PLpgSQL_execstate *estate, PLpgSQL_function *fun
 		/* finalize profile - get result profile */
 
 		profiler_stmt_walker(pinfo, PLPGSQL_CHECK_STMT_WALKER_COUNT_EXEC_TIME,
-							 (PLpgSQL_stmt *) pinfo->function->action, NULL, NULL, 1,
+							 (PLpgSQL_stmt *) pinfo->func->action, NULL, NULL, 1,
 							 &opts);
 
 		update_persistent_profile(pinfo, func);
@@ -2790,6 +2927,66 @@ plpgsql_check_profiler_stmt_beg(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt)
 {
 	profiler_info *pinfo = (profiler_info *) estate->plugin_info;
 
+	if (top_pinfo && top_pinfo->pinfo)
+	{
+		if (estate->eval_econtext != curr_eval_econtext)
+		{
+			if (estate->cur_error)
+			{
+				int		i;
+				bool	found = false;
+
+				/*
+				 * detected exception handler. We have to close
+				 * all statements from stack, until we find common
+				 * eval_context. The reduction is possible, only when
+				 * we found common eval_context. We can lost if if this
+				 * shared context is over then NESTED_STMTS_STACK_SIZE.
+				 */
+				for (i = top_pinfo->nested_stmts_count - 1; i >= 0; i--)
+				{
+					if (i < NESTED_STMTS_STACK_SIZE)
+					{
+						if (top_pinfo->eval_econtext[i] == estate->eval_econtext)
+						{
+							found = true;
+							break;
+						}
+					}
+				}
+
+				if (found)
+				{
+					for (i = top_pinfo->nested_stmts_count - 1; i >= 0; i--)
+					{
+						if (i < NESTED_STMTS_STACK_SIZE)
+						{
+							if (top_pinfo->eval_econtext[i] == estate->eval_econtext)
+							{
+								top_pinfo->nested_stmts_count = i + 1;
+								break;
+							}
+							else
+								plpgsql_check_profiler_stmt_end(NULL, top_pinfo->nested_stmts[i]);
+						}
+					}
+				}
+			}
+
+			curr_eval_econtext = estate->eval_econtext;
+		}
+
+		if (top_pinfo->nested_stmts_count < NESTED_STMTS_STACK_SIZE)
+		{
+			top_pinfo->nested_stmts[top_pinfo->nested_stmts_count] = stmt;
+			top_pinfo->eval_econtext[top_pinfo->nested_stmts_count] = estate->eval_econtext;
+		}
+
+		top_pinfo->nested_stmts_count++;
+		top_pinfo->err_stmt = stmt;
+	}
+
+
 	if (plpgsql_check_tracer && pinfo)
 	{
 
@@ -2838,10 +3035,74 @@ plpgsql_check_profiler_stmt_beg(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt)
 	}
 }
 
+/*
+ * Cleaning mode is used for closing unfinished statements after an exception.
+ */
 void
 plpgsql_check_profiler_stmt_end(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt)
 {
-	profiler_info *pinfo = (profiler_info *) estate->plugin_info;
+	profiler_info *pinfo;
+	bool	cleaning_mode = false;
+	bool	is_error_stmt = false;
+
+	if (!estate)
+	{
+		Assert(top_pinfo && top_pinfo->pinfo);
+
+		pinfo = top_pinfo->pinfo;
+
+
+		estate = pinfo->estate;
+
+		is_error_stmt = top_pinfo->err_stmt == stmt;
+		cleaning_mode = true;
+	}
+	else
+		pinfo = (profiler_info *) estate->plugin_info;
+
+	if (top_pinfo && top_pinfo->pinfo && !cleaning_mode)
+	{
+		int		i;
+		bool	found = false;
+
+		top_pinfo->nested_stmts_count--;
+
+		/*
+		 * try to synchronize counter, we can lost synchronization when
+		 * there was handled exception and deeper statement's stack than
+		 * NESTED_STMTS_STACK_SIZE.
+		 */
+		for (i = top_pinfo->nested_stmts_count; i >= 0; i--)
+		{
+			if (i < NESTED_STMTS_STACK_SIZE)
+			{
+				if (top_pinfo->nested_stmts[i] == stmt)
+				{
+					found = true;
+					break;
+				}
+			}
+		}
+
+		if (found)
+		{
+			for (i = top_pinfo->nested_stmts_count; i >= 0; i--)
+			{
+				if (i < NESTED_STMTS_STACK_SIZE)
+				{
+					if (top_pinfo->nested_stmts[i] == stmt)
+					{
+						top_pinfo->nested_stmts_count = i;
+						break;
+					}
+					else
+						plpgsql_check_profiler_stmt_end(NULL, top_pinfo->nested_stmts[i]);
+				}
+			}
+		}
+
+		top_pinfo->err_stmt = NULL;
+	}
 
 	if (plpgsql_check_tracer && pinfo)
 	{
@@ -2862,12 +3123,14 @@ plpgsql_check_profiler_stmt_end(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt)
 
 #endif
 
-		plpgsql_check_tracer_on_stmt_end(estate, stmt);
+		/* These nodes was not executed */
+		if (!cleaning_mode)
+			plpgsql_check_tracer_on_stmt_end(estate, stmt);
 	}
 
 	if (plpgsql_check_profiler &&
 		pinfo && pinfo->profile &&
-		estate->func->fn_oid != InvalidOid)
+		pinfo->func->fn_oid != InvalidOid)
 	{
 		int stmtid = STMTID(stmt);
 		profiler_stmt *pstmt = &pinfo->stmts[stmtid];
@@ -2877,9 +3140,10 @@ plpgsql_check_profiler_stmt_end(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt)
 
 		Assert(pinfo->pi_magic == PI_MAGIC);
 
-		if (pstmt->queryid == NOQUERYID)
+		if (pstmt->queryid == NOQUERYID && estate)
 			pstmt->queryid = profiler_get_queryid(estate, stmt,
-												  &pstmt->has_queryid);
+												  &pstmt->has_queryid,
+												  &pstmt->qparams);
 
 		INSTR_TIME_SET_CURRENT(end_time);
 		end_time2 = end_time;
@@ -2892,8 +3156,14 @@ plpgsql_check_profiler_stmt_end(PLpgSQL_execstate *estate, PLpgSQL_stmt *stmt)
 			pstmt->us_max = elapsed;
 
 		pstmt->us_total = INSTR_TIME_GET_MICROSEC(pstmt->total);
-		pstmt->rows += estate->eval_processed;
+
+		if (!cleaning_mode)
+			pstmt->rows += estate->eval_processed;
+
 		pstmt->exec_count++;
+
+		if (is_error_stmt)
+			pstmt->exec_count_err++;
 	}
 }
 
@@ -2926,9 +3196,11 @@ plpgsql_check_profiler_iterate_over_all_profiles(plpgsql_check_result_info *ri)
 		Oid		fn_oid,
 				db_oid;
 		uint64	exec_count,
+				exec_count_err,
 				total_time,
 				min_time,
 				max_time;
+
 		float8	total_time_xx;
 		HeapTuple	tp;
 
@@ -2938,6 +3210,7 @@ plpgsql_check_profiler_iterate_over_all_profiles(plpgsql_check_result_info *ri)
 		fn_oid = fstats_item->key.fn_oid;
 		db_oid = fstats_item->key.db_oid;
 		exec_count = fstats_item->exec_count;
+		exec_count_err = fstats_item->exec_count_err;
 		total_time = fstats_item->total_time;
 		total_time_xx = fstats_item->total_time_xx;
 		min_time = fstats_item->min_time;
@@ -2964,13 +3237,110 @@ plpgsql_check_profiler_iterate_over_all_profiles(plpgsql_check_result_info *ri)
 		plpgsql_check_put_profiler_functions_all_tb(ri,
 													fn_oid,
 													exec_count,
-													total_time,
+													exec_count_err,
+													(double) total_time,
 													ceil(total_time / ((double) exec_count)),
 													ceil(sqrt(total_time_xx / exec_count)),
-													min_time,
-													max_time);
+													(double) min_time,
+													(double) max_time);
 	}
 
 	if (htab_is_shared)
 		LWLockRelease(profiler_ss->fstats_lock);
 }
+
+/*
+ * Used as needs_fmgr_hook. All plpgsql functions
+ * needs this hook when profiler is active.
+ */
+bool
+plpgsql_check_needs_fmgr_hook(Oid fn_oid)
+{
+	if (plpgsql_check_next_needs_fmgr_hook &&
+		(*plpgsql_check_next_needs_fmgr_hook)(fn_oid))
+		return true;
+
+	if (!plpgsql_check_profiler)
+		return false;
+
+	return plpgsql_check_is_plpgsql_function(fn_oid);
+}
+
+
+void
+plpgsql_check_fmgr_hook(FmgrHookEventType event,
+						FmgrInfo *flinfo, Datum *private)
+{
+	fmgr_hook_private *stack;
+
+	switch (event)
+	{
+		case FHET_START:
+			stack = (fmgr_hook_private *) DatumGetPointer(*private);
+			if (!stack)
+			{
+				MemoryContext oldcxt;
+
+				oldcxt = MemoryContextSwitchTo(flinfo->fn_mcxt);
+				stack = palloc(sizeof(fmgr_hook_private));
+				stack->use_plpgsql = plpgsql_check_is_plpgsql_function(flinfo->fn_oid);
+				stack->next_private = 0;
+
+				MemoryContextSwitchTo(oldcxt);
+
+				*private = PointerGetDatum(stack);
+			}
+
+			if (stack->use_plpgsql)
+			{
+				profiler_stack *pstack = palloc0(sizeof(profiler_stack));
+
+				pstack->prev_pinfo = top_pinfo;
+				top_pinfo = pstack;
+			}
+
+			if (plpgsql_check_next_fmgr_hook)
+				(*plpgsql_check_next_fmgr_hook) (event, flinfo, &stack->next_private);
+
+			break;
+
+		case FHET_END:
+		case FHET_ABORT:
+			stack = (fmgr_hook_private *) DatumGetPointer(*private);
+
+			/*
+			 * When plpgsql_check is loaded due dependency on some executed function,
+			 * then FHET_START and FHET_END (or FHET_ABORT) can be assymetric. So
+			 * stack can be NULL. See issue #101
+			 */
+			if (stack && stack->use_plpgsql)
+			{
+				profiler_stack *pstack = top_pinfo->prev_pinfo;
+
+				if (event == FHET_ABORT)
+				{
+					profiler_info *pinfo = top_pinfo->pinfo;
+					int		i;
+
+					if (pinfo)
+					{
+						for (i = top_pinfo->nested_stmts_count - 1; i >= 0; i--)
+						{
+							if (i < NESTED_STMTS_STACK_SIZE)
+								plpgsql_check_profiler_stmt_end(NULL, top_pinfo->nested_stmts[i]);
+						}
+
+						plpgsql_check_profiler_func_end(NULL, pinfo->func);
+					}
+				}
+
+				pfree(top_pinfo);
+				top_pinfo = pstack;
+			}
+
+			if (plpgsql_check_next_fmgr_hook)
+				(*plpgsql_check_next_fmgr_hook) (event, flinfo, &stack->next_private);
+
+			break;
+	}
+}
