@@ -14,20 +14,12 @@
 #include "access/tupconvert.h"
 #include "catalog/pg_collation.h"
 #include "catalog/pg_type.h"
-
 #include "nodes/nodeFuncs.h"
+#include "nodes/value.h"
 #include "parser/parse_node.h"
-
-#if PG_VERSION_NUM < 90600
-
-#include "parser/keywords.h"
-
-#else
-
+#include "parser/parser.h"
 #include "common/keywords.h"
 
-#endif
-
 static void check_stmts(PLpgSQL_checkstate *cstate, List *stmts, int *closing, List **exceptions);
 static PLpgSQL_stmt_stack_item * push_stmt_to_stmt_stack(PLpgSQL_checkstate *cstate);
 static void pop_stmt_from_stmt_stack(PLpgSQL_checkstate *cstate);
@@ -136,7 +128,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 {
 	TupleDesc	tupdesc = NULL;
 	PLpgSQL_function *func;
-	ListCell   *l;
 	ResourceOwner oldowner;
 	MemoryContext oldCxt = CurrentMemoryContext;
 	PLpgSQL_stmt_stack_item *outer_stmt;
@@ -297,7 +288,8 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 						if (*closing == PLPGSQL_CHECK_CLOSED_BY_EXCEPTIONS)
 						{
 							ListCell   *lc;
-							int			i = 0;
+							ListCell   *l;
+							int		errn = 0;
 							int    *err_codes = NULL;
 							int		nerr_codes = 0;
 
@@ -307,7 +299,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 							foreach(lc, *exceptions)
 							{
-								err_codes[i++] = lfirst_int(lc);
+								err_codes[errn++] = lfirst_int(lc);
 							}
 
 							foreach(l, stmt_block->exceptions->exc_list)
@@ -320,11 +312,11 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 								if (*exceptions != NIL)
 								{
-									int		i;
+									int		idx;
 
-									for (i = 0; i < nerr_codes; i++)
+									for (idx = 0; idx < nerr_codes; idx++)
 									{
-										int		err_code = err_codes[i];
+										int		err_code = err_codes[idx];
 
 										if (err_code != -1 &&
 											exception_matches_conditions(err_code, exception->conditions))
@@ -333,7 +325,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 																			 &exceptions_transformed, exceptions_local,
 																			 err_code);
 											*exceptions = list_delete_int(*exceptions, err_code);
-											err_codes[i] = -1;
+											err_codes[idx] = -1;
 										}
 									}
 								}
@@ -353,6 +345,8 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 						}
 						else
 						{
+							ListCell   *l;
+
 							closing_handlers = *closing;
 
 							foreach(l, stmt_block->exceptions->exc_list)
@@ -516,8 +510,9 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 				{
 					PLpgSQL_stmt_case *stmt_case = (PLpgSQL_stmt_case *) stmt;
 					Oid			result_oid;
-					int		closing_local;
-					List	*exceptions_local;
+					int			closing_local;
+					List	    *exceptions_local;
+					ListCell    *l;
 					int		closing_all_paths = PLPGSQL_CHECK_UNKNOWN;
 
 					if (stmt_case->t_expr != NULL)
@@ -566,6 +561,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 						ReleaseTupleDesc(tupdesc);
 					}
+
 					foreach(l, stmt_case->case_when_list)
 					{
 						PLpgSQL_case_when *cwt = (PLpgSQL_case_when *) lfirst(l);
@@ -899,7 +895,6 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 					{
 						PLpgSQL_datum *retvar = cstate->estate->datums[stmt_rn->retvarno];
 						PLpgSQL_execstate *estate = cstate->estate;
-						TupleDesc	tupdesc;
 						int		natts;
 
 						cstate->used_variables = bms_add_member(cstate->used_variables, stmt_rn->retvarno);
@@ -1049,6 +1044,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 				{
 					PLpgSQL_stmt_raise *stmt_raise = (PLpgSQL_stmt_raise *) stmt;
 					ListCell   *current_param;
+					ListCell   *l;
 					char	   *cp;
 					int			err_code = 0;
 
@@ -1193,11 +1189,12 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 				{
 					PLpgSQL_stmt_open *stmt_open = (PLpgSQL_stmt_open *) stmt;
 					PLpgSQL_var *var = (PLpgSQL_var *) (cstate->estate->datums[stmt_open->curvar]);
+					ListCell	*l;
 
 					plpgsql_check_expr_as_sqlstmt_data(cstate, var->cursor_explicit_expr);
 					plpgsql_check_expr_as_sqlstmt_data(cstate, stmt_open->query);
 
-					if (var != NULL && stmt_open->query != NULL)
+					if (stmt_open->query != NULL)
 						var->cursor_explicit_expr = stmt_open->query;
 
 					plpgsql_check_expr_as_sqlstmt_data(cstate, stmt_open->argquery);
@@ -1265,12 +1262,14 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 #if PG_VERSION_NUM >= 110000
 
+#if PG_VERSION_NUM < 140000
 			case PLPGSQL_STMT_SET:
 				/*
 				 * We can not check this now, syntax should be ok.
 				 * The expression there has not plan.
 				 */
 				break;
+#endif			/* PG_VERSION_NUM < 140000 */
 
 			case PLPGSQL_STMT_COMMIT:
 			case PLPGSQL_STMT_ROLLBACK:
@@ -1307,7 +1306,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 				}
 				break;
 
-#endif
+#endif			/* PG_VERSION_NUM >= 110000 */
 
 			default:
 				elog(ERROR, "unrecognized cmd_type: %d", stmt->cmd_type);
@@ -1315,7 +1314,7 @@ plpgsql_check_stmt(PLpgSQL_checkstate *cstate, PLpgSQL_stmt *stmt, int *closing,
 
 		pop_stmt_from_stmt_stack(cstate);
 
-		RollbackAndReleaseCurrentSubTransaction();
+		ReleaseCurrentSubTransaction();
 		MemoryContextSwitchTo(oldCxt);
 		CurrentResourceOwner = oldowner;
 
@@ -1371,7 +1370,7 @@ check_stmts(PLpgSQL_checkstate *cstate, List *stmts, int *closing, List **except
 	ListCell   *lc;
 	int			closing_local;
 	List	   *exceptions_local;
-	bool		dead_code_alert = false;
+	volatile bool		dead_code_alert = false;
 	plpgsql_check_pragma_vector		prev_pragma_vector = cstate->pragma_vector;
 
 	*closing = PLPGSQL_CHECK_UNCLOSED;
@@ -1688,15 +1687,8 @@ exception_matches_conditions(int sqlerrstate, PLpgSQL_condition *cond)
 		 */
 		if (_sqlerrstate == 0)
 		{
-			if (sqlerrstate != ERRCODE_QUERY_CANCELED
-
-#if PG_VERSION_NUM >= 90500
-
-				 && sqlerrstate != ERRCODE_ASSERT_FAILURE
-
-#endif
-
-				)
+			if (sqlerrstate != ERRCODE_QUERY_CANCELED &&
+				 sqlerrstate != ERRCODE_ASSERT_FAILURE)
 				return true;
 		}
 		/* Exact match? */
@@ -1823,11 +1815,15 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 
 				  List *params)
 {
-	bool		prev_has_execute_stmt = cstate->has_execute_stmt;
 	Node	   *expr_node;
 	ListCell   *l;
 	int			loc = -1;
+	char	   *dynquery = NULL;
+	bool		prev_has_execute_stmt = cstate->has_execute_stmt;
+	volatile bool expr_is_const = false;
+
 	volatile bool raise_unknown_rec_warning = false;
+	volatile bool known_type_of_dynexpr = false;
 
 	/*
 	 * possible checks:
@@ -1839,6 +1835,10 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 	 * 2. When expression is real expression, then we should to check any string
 	 *    kind parameters if are sanitized by functions quote_ident, qoute_literal,
 	 *    or format.
+	 *
+	 * 3. When expression is based on calling format function, and there are used
+	 *    only placeholders %I and %L, then we can try to check syntax of embeded
+	 *    query.
 	 */
 
 	cstate->has_execute_stmt = true;
@@ -1851,12 +1851,110 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 	plpgsql_check_expr(cstate, query);
 	expr_node = plpgsql_check_expr_get_node(cstate, query, false);
 
-	if (IsA(expr_node, Const))
+	if (IsA(expr_node, FuncExpr))
 	{
-		char *query = plpgsql_check_const_to_string((Const *) expr_node);
-		PLpgSQL_expr	   *dynexpr;
-		DynSQLParams		dsp;
-		bool				is_mp;
+		FuncExpr *fexpr = (FuncExpr *) expr_node;
+
+		if (fexpr->funcid == FORMAT_0PARAM_OID ||
+			fexpr->funcid == FORMAT_NPARAM_OID)
+		{
+			if (fexpr->args && IsA(linitial(fexpr->args), Const))
+			{
+				StringInfoData	sinfo;
+				char		c, *fmt;
+				bool		subst_is_ok = true;
+				bool		found_ident_placeholder = false;
+				bool		found_literal_placeholder = false;
+
+				expr_is_const = fexpr->funcid == FORMAT_0PARAM_OID;
+				fmt = plpgsql_check_const_to_string((Const *) linitial(fexpr->args));
+
+				/*
+				 * The placeholders can be used only in FORMAT_NPARAM function,
+				 * but for simplicity and consistency we check FORMAT_0PARAM and
+				 * FORMAT_NPARAM together
+				 */
+				initStringInfo(&sinfo);
+
+				while ((c = *fmt++))
+				{
+					if (c == '%')
+					{
+						c = *fmt++;
+
+						if (c == '%')
+						{
+							appendStringInfoChar(&sinfo, c);
+						}
+						else if (c == 'I')
+						{
+							appendStringInfoString(&sinfo, "\"%I\"");
+							expr_is_const = false;
+							found_ident_placeholder = true;
+						}
+						else if (c == 'L')
+						{
+							/*
+							 * Original idea was used external parameter,
+							 * but external parameters requires known type,
+							 * so most safe value is NULL instead.
+							 */
+							appendStringInfo(&sinfo, " null ");
+							found_literal_placeholder = false;
+							expr_is_const = false;
+						}
+						else
+						{
+							/*
+							 * Because %s is used, we know nothing about form
+							 * of output string, and has not any sense to continue
+							 * in check.
+							 */
+							subst_is_ok = false;
+							expr_is_const = false;
+							break;
+						}
+					}
+					else
+						appendStringInfoChar(&sinfo, c);
+				}
+
+				if (subst_is_ok)
+				{
+					if (!found_literal_placeholder)
+					{
+
+#if PG_VERSION_NUM >= 140000
+
+						/* in this case we can do only basic parser check */
+						raw_parser(sinfo.data, RAW_PARSE_DEFAULT);
+
+#else
+
+						raw_parser(sinfo.data);
+
+#endif
+
+					}
+
+					if (!found_ident_placeholder)
+						dynquery = sinfo.data;
+				}
+			}
+		}
+	}
+	else if (IsA(expr_node, Const))
+	{
+		expr_is_const = true;
+		dynquery = plpgsql_check_const_to_string((Const *) expr_node);
+	}
+
+	if (dynquery)
+	{
+		PLpgSQL_expr *dynexpr = NULL;
+		DynSQLParams dsp;
+		volatile bool		is_mp;
+		volatile bool is_ok = true;
 
 		dynexpr = palloc0(sizeof(PLpgSQL_expr));
 
@@ -1871,42 +1969,101 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 
 		dynexpr->expr_rw_param = NULL;
 
-#elif PG_VERSION_NUM >= 90500
+#else
 
 		dynexpr->rwparam = -1;
 
 #endif
 
-
-
-		dynexpr->query = query;
+		dynexpr->query = dynquery;
 
 		dsp.args = params;
 		dsp.cstate = cstate;
 		dsp.use_params = false;
 
-		PG_TRY();
+		/*
+		 * When dynquery is not really constant, then there are
+		 * possible false alarms because we try to replace string
+		 * literal by parameter, so we can use it just for type
+		 * detection when check is ok.
+		 */
+		if (expr_is_const)
 		{
-			cstate->allow_mp = true;
+			PG_TRY();
+			{
+				cstate->allow_mp = true;
 
-			plpgsql_check_expr_generic_with_parser_setup(cstate,
+				plpgsql_check_expr_generic_with_parser_setup(cstate,
 													 dynexpr,
 													 (ParserSetupHook) dynsql_parser_setup,
 													 &dsp);
 
-			is_mp = cstate->has_mp;
-			cstate->has_mp = false;
+				is_mp = cstate->has_mp;
+				cstate->has_mp = false;
+			}
+			PG_CATCH();
+			{
+				cstate->allow_mp = false;
+				cstate->has_mp = false;
+
+				PG_RE_THROW();
+			}
+			PG_END_TRY();
 		}
-		PG_CATCH();
+		else
 		{
-			cstate->allow_mp = false;
-			cstate->has_mp = false;
+			MemoryContext oldCxt;
+			ResourceOwner oldowner;
 
-			PG_RE_THROW();
+			/*
+			 * When dynquery is not really constant, then there are
+			 * possible false alarms because we try to replace string
+			 * literal by parameter, so we can use it just for type
+			 * detection when check is ok.
+			 */
+
+			oldCxt = CurrentMemoryContext;
+
+			oldowner = CurrentResourceOwner;
+			BeginInternalSubTransaction(NULL);
+			MemoryContextSwitchTo(cstate->check_cxt);
+
+			PG_TRY();
+			{
+				cstate->allow_mp = true;
+
+				plpgsql_check_expr_generic_with_parser_setup(cstate,
+													 dynexpr,
+													 (ParserSetupHook) dynsql_parser_setup,
+													 &dsp);
+
+				is_mp = cstate->has_mp;
+				cstate->has_mp = false;
+
+				RollbackAndReleaseCurrentSubTransaction();
+				MemoryContextSwitchTo(oldCxt);
+				CurrentResourceOwner = oldowner;
+
+				SPI_restore_connection();
+			}
+			PG_CATCH();
+			{
+				is_ok = false;
+
+				cstate->allow_mp = false;
+				cstate->has_mp = false;
+
+				MemoryContextSwitchTo(oldCxt);
+				FlushErrorState();
+
+				RollbackAndReleaseCurrentSubTransaction();
+				MemoryContextSwitchTo(oldCxt);
+				CurrentResourceOwner = oldowner;
+			}
+			PG_END_TRY();
 		}
-		PG_END_TRY();
 
-		if (!is_mp && (!params || !dsp.use_params))
+		if (is_ok && expr_is_const && !is_mp && (!params || !dsp.use_params))
 		{
 
 			/* probably useless dynamic command */
@@ -1919,7 +2076,7 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 									0, NULL, NULL);
 		}
 
-		if (params && !dsp.use_params)
+		if (is_ok && params && !dsp.use_params)
 		{
 			plpgsql_check_put_error(cstate,
 									0, 0,
@@ -1930,8 +2087,10 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 									0, NULL, NULL);
 		}
 
-		if (dynexpr->plan)
+		if (is_ok && dynexpr->plan)
 		{
+			known_type_of_dynexpr = true;
+
 			if (stmt->cmd_type == PLPGSQL_STMT_RETURN_QUERY)
 			{
 				plpgsql_check_returned_expr(cstate, dynexpr, false);
@@ -1959,7 +2118,8 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 		if (!is_mp)
 			cstate->has_execute_stmt = prev_has_execute_stmt;
 	}
-	else
+
+	if (!expr_is_const)
 	{
 		/*
 		 * execute string is not constant (is not safe),
@@ -1991,7 +2151,8 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 		}
 
 		/* in this case we don't know number of output columns */
-		if (stmt->cmd_type == PLPGSQL_STMT_RETURN_QUERY)
+		if (stmt->cmd_type == PLPGSQL_STMT_RETURN_QUERY &&
+			!known_type_of_dynexpr)
 		{
 			cstate->found_return_dyn_query = true;
 		}
@@ -2000,7 +2161,7 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 		 * In this case, we don't know a result type, and we should
 		 * to raise warning about this situation.
 		 */
-		if (into)
+		if (into && !known_type_of_dynexpr)
 		{
 
 #if PG_VERSION_NUM >= 110000
@@ -2039,13 +2200,24 @@ check_dynamic_sql(PLpgSQL_checkstate *cstate,
 #endif
 
 		{
-			plpgsql_check_put_error(cstate,
-									0, 0,
-									"cannot determinate a result of dynamic SQL",
-									"There is a risk of related false alarms.",
-						  "Don't use dynamic SQL and record type together, when you would check function.",
-									PLPGSQL_CHECK_WARNING_OTHERS,
-									0, NULL, NULL);
+
+#if PG_VERSION_NUM >= 110000
+
+			if (!bms_is_member(target->dno, cstate->typed_variables))
+
+#else
+
+			if (!bms_is_member(rec->dno, cstate->typed_variables))
+
+#endif
+
+				plpgsql_check_put_error(cstate,
+										0, 0,
+										"cannot determinate a result of dynamic SQL",
+										"There is a risk of related false alarms.",
+							  "Don't use dynamic SQL and record type together, when you would check function.",
+										PLPGSQL_CHECK_WARNING_OTHERS,
+										0, NULL, NULL);
 		}
 	}
 }
