@@ -4,7 +4,7 @@
  *
  *			  top functions - display results in table format
  *
- * by Pavel Stehule 2013-2020
+ * by Pavel Stehule 2013-2021
  *
  *-------------------------------------------------------------------------
  */
@@ -27,6 +27,7 @@ PG_FUNCTION_INFO_V1(plpgsql_check_function_tb_name);
 PG_FUNCTION_INFO_V1(plpgsql_show_dependency_tb_name);
 PG_FUNCTION_INFO_V1(plpgsql_profiler_function_tb_name);
 PG_FUNCTION_INFO_V1(plpgsql_profiler_function_statements_tb_name);
+PG_FUNCTION_INFO_V1(plpgsql_profiler_functions_all_tb);
 
 #define ERR_NULL_OPTION(option)		ereport(ERROR, \
 									  (errcode(ERRCODE_NULL_VALUE_NOT_ALLOWED), \
@@ -457,7 +458,9 @@ profiler_function_statements_tb_internal(Oid fnoid, FunctionCallInfo fcinfo)
 
 	plpgsql_check_init_ri(&ri, PLPGSQL_SHOW_PROFILE_STATEMENTS_TABULAR, rsinfo);
 
-	plpgsql_check_profiler_show_profile_statements(&ri, &cinfo, NULL);
+	plpgsql_check_iterate_over_profile(&cinfo,
+									   PLPGSQL_CHECK_STMT_WALKER_PREPARE_RESULT,
+									   &ri, NULL);
 
 	plpgsql_check_finalize_ri(&ri);
 
@@ -608,3 +611,20 @@ plpgsql_profiler_function_statements_tb_name(PG_FUNCTION_ARGS)
 
 	return profiler_function_statements_tb_internal(fnoid, fcinfo);
 }
+
+Datum
+plpgsql_profiler_functions_all_tb(PG_FUNCTION_ARGS)
+{
+	plpgsql_check_result_info ri;
+	ReturnSetInfo *rsinfo;
+
+	/* check to see if caller supports us returning a tuplestore */
+	rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
+	SetReturningFunctionCheck(rsinfo);
+
+	plpgsql_check_init_ri(&ri, PLPGSQL_SHOW_PROFILE_FUNCTIONS_ALL_TABULAR, rsinfo);
+
+	plpgsql_check_profiler_iterate_over_all_profiles(&ri);
+
+	return (Datum) 0;
+}
