@@ -75,6 +75,9 @@ plpgsql_check_CallExprGetRowTarget(PLpgSQL_checkstate *cstate, PLpgSQL_expr *Cal
 
 		/* Extract function arguments, and expand any named-arg notation */
 		funcargs = expand_function_arguments(funcexpr->args,
+#if PG_VERSION_NUM >= 140000
+											 true,
+#endif
 											 funcexpr->funcresulttype,
 											 tuple);
 
@@ -516,13 +519,9 @@ plpgsql_check_expr_get_desc(PLpgSQL_checkstate *cstate,
 
 	cplan = GetCachedPlan(plansource, NULL, NULL, NULL);
 
-#elif PG_VERSION_NUM >= 100000
-
-	cplan = GetCachedPlan(plansource, NULL, true, NULL);
-
 #else
 
-	cplan = GetCachedPlan(plansource, NULL, true);
+	cplan = GetCachedPlan(plansource, NULL, true, NULL);
 
 #endif
 		_stmt = (PlannedStmt *) linitial(cplan->stmt_list);
@@ -631,7 +630,16 @@ plpgsql_check_expr_get_desc(PLpgSQL_checkstate *cstate,
 
 							FreeTupleDesc(tupdesc);
 
+#if PG_VERSION_NUM >= 140000
+
+							if (c->consttype == RECORDOID && c->consttypmod == -1 && !c->constisnull)
+
+#else
+
 							if (c->consttype == RECORDOID && c->consttypmod == -1)
+
+#endif
+
 							{
 								Oid		tupType;
 								int32	tupTypmod;
